"""
    tests.support.zfs
    ~~~~~~~~~~~~~~~~~

    ZFS related unit test data structures
"""


import salt.utils.zfs
from tests.support.mock import MagicMock, patch


class ZFSMockData:
    def __init__(self):
        # property_map mocks
        self.pmap_exec_zpool = {
            "retcode": 2,
            "stdout": "",
            "stderr": "\n".join(
                [
                    "missing property argument",
                    "usage:",
                    '        get [-Hp] [-o "all" | field[,...]] <"all" |'
                    " property[,...]> <pool> ...",
                    "",
                    "the following properties are supported:",
                    "",
                    "        PROPERTY         EDIT   VALUES",
                    "",
                    "        allocated          NO   <size>",
                    "        capacity           NO   <size>",
                    "        dedupratio         NO   <1.00x or higher if deduped>",
                    "        expandsize         NO   <size>",
                    "        fragmentation      NO   <percent>",
                    "        free               NO   <size>",
                    "        freeing            NO   <size>",
                    "        guid               NO   <guid>",
                    "        health             NO   <state>",
                    "        leaked             NO   <size>",
                    "        size               NO   <size>",
                    "        altroot           YES   <path>",
                    "        autoexpand        YES   on | off",
                    "        autoreplace       YES   on | off",
                    "        bootfs            YES   <filesystem>",
                    "        bootsize          YES   <size>",
                    "        cachefile         YES   <file> | none",
                    "        comment           YES   <comment-string>",
                    "        dedupditto        YES   <threshold (min 100)>",
                    "        delegation        YES   on | off",
                    "        failmode          YES   wait | continue | panic",
                    "        listsnapshots     YES   on | off",
                    "        readonly          YES   on | off",
                    "        version           YES   <version>",
                    "        feature@...       YES   disabled | enabled | active",
                    "",
                    "The feature@ properties must be appended with a feature name.",
                    "See zpool-features(5). ",
                ]
            ),
        }
        self.pmap_zpool = {
            "comment": {"edit": True, "type": "str", "values": "<comment-string>"},
            "freeing": {"edit": False, "type": "size", "values": "<size>"},
            "listsnapshots": {"edit": True, "type": "bool", "values": "on | off"},
            "leaked": {"edit": False, "type": "size", "values": "<size>"},
            "version": {"edit": True, "type": "numeric", "values": "<version>"},
            "write": {"edit": False, "type": "size", "values": "<size>"},
            "replace": {"edit": True, "type": "bool", "values": "on | off"},
            "delegation": {"edit": True, "type": "bool", "values": "on | off"},
            "dedupditto": {
                "edit": True,
                "type": "str",
                "values": "<threshold (min 100)>",
            },
            "autoexpand": {"edit": True, "type": "bool", "values": "on | off"},
            "alloc": {"edit": False, "type": "size", "values": "<size>"},
            "allocated": {"edit": False, "type": "size", "values": "<size>"},
            "guid": {"edit": False, "type": "numeric", "values": "<guid>"},
            "size": {"edit": False, "type": "size", "values": "<size>"},
            "cap": {"edit": False, "type": "numeric", "values": "<count>"},
            "capacity": {"edit": False, "type": "size", "values": "<size>"},
            "capacity-alloc": {"edit": False, "type": "size", "values": "<size>"},
            "capacity-free": {"edit": False, "type": "size", "values": "<size>"},
            "cachefile": {"edit": True, "type": "str", "values": "<file> | none"},
            "cksum": {"edit": False, "type": "numeric", "values": "<count>"},
            "bootfs": {"edit": True, "type": "str", "values": "<filesystem>"},
            "autoreplace": {"edit": True, "type": "bool", "values": "on | off"},
            "bandwidth-read": {"edit": False, "type": "size", "values": "<size>"},
            "bandwidth-write": {"edit": False, "type": "size", "values": "<size>"},
            "operations-read": {"edit": False, "type": "size", "values": "<size>"},
            "operations-write": {"edit": False, "type": "size", "values": "<size>"},
            "read": {"edit": False, "type": "size", "values": "<size>"},
            "readonly": {"edit": True, "type": "bool", "values": "on | off"},
            "dedupratio": {
                "edit": False,
                "type": "str",
                "values": "<1.00x or higher if deduped>",
            },
            "health": {"edit": False, "type": "str", "values": "<state>"},
            "feature@": {
                "edit": True,
                "type": "str",
                "values": "disabled | enabled | active",
            },
            "expandsize": {"edit": False, "type": "size", "values": "<size>"},
            "listsnaps": {"edit": True, "type": "bool", "values": "on | off"},
            "bootsize": {"edit": True, "type": "size", "values": "<size>"},
            "free": {"edit": False, "type": "size", "values": "<size>"},
            "failmode": {
                "edit": True,
                "type": "str",
                "values": "wait | continue | panic",
            },
            "altroot": {"edit": True, "type": "str", "values": "<path>"},
            "expand": {"edit": True, "type": "bool", "values": "on | off"},
            "frag": {"edit": False, "type": "str", "values": "<percent>"},
            "fragmentation": {"edit": False, "type": "str", "values": "<percent>"},
        }

        # fmt: off
        self.pmap_exec_zfs = {
            "retcode": 2,
            "stdout": "",
            "stderr": "\n".join(
                [
                    "missing property argument",
                    "usage:",
                    '        get [-crHp] [-d max] [-o "all" | field[,...]]',
                    "            [-t type[,...]] [-s source[,...]]",
                    '            <"all" | property[,...]>'
                    " [filesystem|volume|snapshot|bookmark] ...",
                    "",
                    "The following properties are supported:",
                    "",
                    "        PROPERTY       EDIT  INHERIT   VALUES",
                    "",
                    "        available        NO       NO   <size>",
                    "        clones           NO       NO   <dataset>[,...]",
                    "        compressratio    NO       NO   <1.00x or higher if compressed>",
                    "        creation         NO       NO   <date>",
                    "        defer_destroy    NO       NO   yes | no",
                    "        filesystem_count  NO       NO   <count>",
                    "        logicalreferenced  NO       NO   <size>",
                    "        logicalused      NO       NO   <size>",
                    "        mounted          NO       NO   yes | no",
                    "        origin           NO       NO   <snapshot>",
                    "        receive_resume_token  NO       NO   <string token>",
                    "        refcompressratio  NO       NO   <1.00x or higher if compressed>",
                    "        referenced       NO       NO   <size>",
                    "        snapshot_count   NO       NO   <count>",
                    "        type             NO       NO   filesystem | volume | snapshot | bookmark",
                    "        used             NO       NO   <size>",
                    "        usedbychildren   NO       NO   <size>",
                    "        usedbydataset    NO       NO   <size>",
                    "        usedbyrefreservation  NO       NO   <size>",
                    "        usedbysnapshots  NO       NO   <size>",
                    "        userrefs         NO       NO   <count>",
                    "        written          NO       NO   <size>",
                    "        aclinherit      YES      YES   discard | noallow | restricted | passthrough | passthrough-x",
                    "        aclmode         YES      YES   discard | groupmask | passthrough | restricted",
                    "        atime           YES      YES   on | off",
                    "        canmount        YES       NO   on | off | noauto",
                    "        casesensitivity  NO      YES   sensitive | insensitive | mixed",
                    "        checksum        YES      YES   on | off | fletcher2 | fletcher4 | sha256 | sha512 | skein | edonr",
                    "        compression     YES      YES   on | off | lzjb | gzip | gzip-[1-9] | zle | lz4",
                    "        copies          YES      YES   1 | 2 | 3",
                    "        dedup           YES      YES   on | off | verify | sha256[,verify], sha512[,verify], skein[,verify], edonr,verify",
                    "        devices         YES      YES   on | off",
                    "        exec            YES      YES   on | off",
                    "        filesystem_limit YES       NO   <count> | none",
                    "        logbias         YES      YES   latency | throughput",
                    "        mlslabel        YES      YES   <sensitivity label>",
                    "        mountpoint      YES      YES   <path> | legacy | none",
                    "        nbmand          YES      YES   on | off",
                    "        normalization    NO      YES   none | formC | formD | formKC | formKD",
                    "        primarycache    YES      YES   all | none | metadata",
                    "        quota           YES       NO   <size> | none",
                    "        readonly        YES      YES   on | off",
                    "        recordsize      YES      YES   512 to 1M, power of 2",
                    "        redundant_metadata YES      YES   all | most",
                    "        refquota        YES       NO   <size> | none",
                    "        refreservation  YES       NO   <size> | none",
                    "        reservation     YES       NO   <size> | none",
                    "        secondarycache  YES      YES   all | none | metadata",
                    "        setuid          YES      YES   on | off",
                    "        sharenfs        YES      YES   on | off | share(1M) options",
                    "        sharesmb        YES      YES   on | off | sharemgr(1M) options",
                    "        snapdir         YES      YES   hidden | visible",
                    "        snapshot_limit  YES       NO   <count> | none",
                    "        sync            YES      YES   standard | always | disabled",
                    "        utf8only         NO      YES   on | off",
                    "        version         YES       NO   1 | 2 | 3 | 4 | 5 | current",
                    "        volblocksize     NO      YES   512 to 128k, power of 2",
                    "        volsize         YES       NO   <size>",
                    "        vscan           YES      YES   on | off",
                    "        xattr           YES      YES   on | off",
                    "        zoned           YES      YES   on | off",
                    "        userused@...     NO       NO   <size>",
                    "        groupused@...    NO       NO   <size>",
                    "        userquota@...   YES       NO   <size> | none",
                    "        groupquota@...  YES       NO   <size> | none",
                    "        written@<snap>   NO       NO   <size>",
                    "",
                    "Sizes are specified in bytes with standard units such as K, M, G, etc.",
                    "",
                    "User-defined properties can be specified by using a name containing a colon (:).",
                    "",
                    "The {user|group}{used|quota}@ properties must be appended with",
                    "a user or group specifier of one of these forms:",
                    '    POSIX name      (eg: "matt")',
                    '    POSIX id        (eg: "126829")',
                    '    SMB name@domain (eg: "matt@sun")',
                    '    SMB SID         (eg: "S-1-234-567-89")',
                ]
            ),
        }
        # fmt: on
        self.pmap_zfs = {
            "origin": {
                "edit": False,
                "inherit": False,
                "values": "<snapshot>",
                "type": "str",
            },
            "setuid": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "referenced": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "vscan": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "logicalused": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "userrefs": {
                "edit": False,
                "inherit": False,
                "values": "<count>",
                "type": "numeric",
            },
            "primarycache": {
                "edit": True,
                "inherit": True,
                "values": "all | none | metadata",
                "type": "str",
            },
            "logbias": {
                "edit": True,
                "inherit": True,
                "values": "latency | throughput",
                "type": "str",
            },
            "creation": {
                "edit": False,
                "inherit": False,
                "values": "<date>",
                "type": "str",
            },
            "sync": {
                "edit": True,
                "inherit": True,
                "values": "standard | always | disabled",
                "type": "str",
            },
            "dedup": {
                "edit": True,
                "inherit": True,
                "values": (
                    "on | off | verify | sha256[,verify], sha512[,verify],"
                    " skein[,verify], edonr,verify"
                ),
                "type": "bool",
            },
            "sharenfs": {
                "edit": True,
                "inherit": True,
                "values": "on | off | share(1m) options",
                "type": "bool",
            },
            "receive_resume_token": {
                "edit": False,
                "inherit": False,
                "values": "<string token>",
                "type": "str",
            },
            "usedbyrefreservation": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "sharesmb": {
                "edit": True,
                "inherit": True,
                "values": "on | off | sharemgr(1m) options",
                "type": "bool",
            },
            "rdonly": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "reservation": {
                "edit": True,
                "inherit": False,
                "values": "<size> | none",
                "type": "size",
            },
            "reserv": {
                "edit": True,
                "inherit": False,
                "values": "<size> | none",
                "type": "size",
            },
            "mountpoint": {
                "edit": True,
                "inherit": True,
                "values": "<path> | legacy | none",
                "type": "str",
            },
            "casesensitivity": {
                "edit": False,
                "inherit": True,
                "values": "sensitive | insensitive | mixed",
                "type": "str",
            },
            "utf8only": {
                "edit": False,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "usedbysnapshots": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "readonly": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "written@": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "avail": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "recsize": {
                "edit": True,
                "inherit": True,
                "values": "512 to 1m, power of 2",
                "type": "str",
            },
            "atime": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "compression": {
                "edit": True,
                "inherit": True,
                "values": "on | off | lzjb | gzip | gzip-[1-9] | zle | lz4",
                "type": "bool",
            },
            "snapdir": {
                "edit": True,
                "inherit": True,
                "values": "hidden | visible",
                "type": "str",
            },
            "aclmode": {
                "edit": True,
                "inherit": True,
                "values": "discard | groupmask | passthrough | restricted",
                "type": "str",
            },
            "zoned": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "copies": {
                "edit": True,
                "inherit": True,
                "values": "1 | 2 | 3",
                "type": "numeric",
            },
            "snapshot_limit": {
                "edit": True,
                "inherit": False,
                "values": "<count> | none",
                "type": "numeric",
            },
            "aclinherit": {
                "edit": True,
                "inherit": True,
                "values": (
                    "discard | noallow | restricted | passthrough | passthrough-x"
                ),
                "type": "str",
            },
            "compressratio": {
                "edit": False,
                "inherit": False,
                "values": "<1.00x or higher if compressed>",
                "type": "str",
            },
            "xattr": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "written": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "version": {
                "edit": True,
                "inherit": False,
                "values": "1 | 2 | 3 | 4 | 5 | current",
                "type": "numeric",
            },
            "recordsize": {
                "edit": True,
                "inherit": True,
                "values": "512 to 1m, power of 2",
                "type": "str",
            },
            "refquota": {
                "edit": True,
                "inherit": False,
                "values": "<size> | none",
                "type": "size",
            },
            "filesystem_limit": {
                "edit": True,
                "inherit": False,
                "values": "<count> | none",
                "type": "numeric",
            },
            "lrefer.": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "type": {
                "edit": False,
                "inherit": False,
                "values": "filesystem | volume | snapshot | bookmark",
                "type": "str",
            },
            "secondarycache": {
                "edit": True,
                "inherit": True,
                "values": "all | none | metadata",
                "type": "str",
            },
            "refer": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "available": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "used": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "exec": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "compress": {
                "edit": True,
                "inherit": True,
                "values": "on | off | lzjb | gzip | gzip-[1-9] | zle | lz4",
                "type": "bool",
            },
            "volblock": {
                "edit": False,
                "inherit": True,
                "values": "512 to 128k, power of 2",
                "type": "str",
            },
            "refcompressratio": {
                "edit": False,
                "inherit": False,
                "values": "<1.00x or higher if compressed>",
                "type": "str",
            },
            "quota": {
                "edit": True,
                "inherit": False,
                "values": "<size> | none",
                "type": "size",
            },
            "groupquota@": {
                "edit": True,
                "inherit": False,
                "values": "<size> | none",
                "type": "size",
            },
            "userquota@": {
                "edit": True,
                "inherit": False,
                "values": "<size> | none",
                "type": "size",
            },
            "snapshot_count": {
                "edit": False,
                "inherit": False,
                "values": "<count>",
                "type": "numeric",
            },
            "volsize": {
                "edit": True,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "clones": {
                "edit": False,
                "inherit": False,
                "values": "<dataset>[,...]",
                "type": "str",
            },
            "canmount": {
                "edit": True,
                "inherit": False,
                "values": "on | off | noauto",
                "type": "bool",
            },
            "mounted": {
                "edit": False,
                "inherit": False,
                "values": "yes | no",
                "type": "bool_alt",
            },
            "groupused@": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "normalization": {
                "edit": False,
                "inherit": True,
                "values": "none | formc | formd | formkc | formkd",
                "type": "str",
            },
            "usedbychildren": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "usedbydataset": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "mlslabel": {
                "edit": True,
                "inherit": True,
                "values": "<sensitivity label>",
                "type": "str",
            },
            "refreserv": {
                "edit": True,
                "inherit": False,
                "values": "<size> | none",
                "type": "size",
            },
            "defer_destroy": {
                "edit": False,
                "inherit": False,
                "values": "yes | no",
                "type": "bool_alt",
            },
            "volblocksize": {
                "edit": False,
                "inherit": True,
                "values": "512 to 128k, power of 2",
                "type": "str",
            },
            "lused.": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "redundant_metadata": {
                "edit": True,
                "inherit": True,
                "values": "all | most",
                "type": "str",
            },
            "filesystem_count": {
                "edit": False,
                "inherit": False,
                "values": "<count>",
                "type": "numeric",
            },
            "devices": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
            "refreservation": {
                "edit": True,
                "inherit": False,
                "values": "<size> | none",
                "type": "size",
            },
            "userused@": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "logicalreferenced": {
                "edit": False,
                "inherit": False,
                "values": "<size>",
                "type": "size",
            },
            "checksum": {
                "edit": True,
                "inherit": True,
                "values": (
                    "on | off | fletcher2 | fletcher4 | sha256 | sha512 | skein | edonr"
                ),
                "type": "bool",
            },
            "nbmand": {
                "edit": True,
                "inherit": True,
                "values": "on | off",
                "type": "bool",
            },
        }

    def _from_auto(self, name, value, source="auto"):
        """
        some more complex patching for zfs.from_auto
        """
        with patch.object(
            salt.utils.zfs,
            "property_data_zpool",
            MagicMock(return_value=self.pmap_zpool),
        ), patch.object(
            salt.utils.zfs, "property_data_zfs", MagicMock(return_value=self.pmap_zfs)
        ):
            return salt.utils.zfs.from_auto(name, value, source)

    def _from_auto_dict(self, values, source="auto"):
        """
        some more complex patching for zfs.from_auto_dict
        """
        with patch.object(
            salt.utils.zfs,
            "property_data_zpool",
            MagicMock(return_value=self.pmap_zpool),
        ), patch.object(
            salt.utils.zfs, "property_data_zfs", MagicMock(return_value=self.pmap_zfs)
        ):
            return salt.utils.zfs.from_auto_dict(values, source)

    def _to_auto(self, name, value, source="auto", convert_to_human=True):
        """
        some more complex patching for zfs.to_auto
        """
        with patch.object(
            salt.utils.zfs,
            "property_data_zpool",
            MagicMock(return_value=self.pmap_zpool),
        ), patch.object(
            salt.utils.zfs, "property_data_zfs", MagicMock(return_value=self.pmap_zfs)
        ):
            return salt.utils.zfs.to_auto(name, value, source, convert_to_human)

    def _to_auto_dict(self, values, source="auto", convert_to_human=True):
        """
        some more complex patching for zfs.to_auto_dict
        """
        with patch.object(
            salt.utils.zfs,
            "property_data_zpool",
            MagicMock(return_value=self.pmap_zpool),
        ), patch.object(
            salt.utils.zfs, "property_data_zfs", MagicMock(return_value=self.pmap_zfs)
        ):
            return salt.utils.zfs.to_auto_dict(values, source, convert_to_human)

    def get_patched_utils(self):
        return {
            "zfs.is_supported": MagicMock(return_value=True),
            "zfs.has_feature_flags": MagicMock(return_value=True),
            "zfs.property_data_zpool": MagicMock(return_value=self.pmap_zpool),
            "zfs.property_data_zfs": MagicMock(return_value=self.pmap_zfs),
            # NOTE: we make zpool_command and zfs_command a NOOP
            #       these are extensively tested in tests.unit.utils.test_zfs
            "zfs.zpool_command": MagicMock(return_value="/bin/false"),
            "zfs.zfs_command": MagicMock(return_value="/bin/false"),
            # NOTE: from_auto_dict is a special snowflake
            #       internally it calls multiple calls from
            #       salt.utils.zfs but we cannot patch those using
            #       the common methode, __utils__ is not available
            #       so they are direct calls, we do some voodoo here.
            "zfs.from_auto_dict": self._from_auto_dict,
            "zfs.from_auto": self._from_auto,
            "zfs.to_auto_dict": self._to_auto_dict,
            "zfs.to_auto": self._to_auto,
        }



import datetime
import functools
import io
import multiprocessing
import os
import signal
import sys
import tempfile
import threading
import time
import warnings

import pytest
import salt.utils.platform
import salt.utils.process
from salt.utils.versions import warn_until_date
from tests.support.mock import patch
from tests.support.unit import TestCase, skipIf

HAS_PSUTIL = False
try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    pass


def die(func):
    """
    Add proc title
    """

    @functools.wraps(func)
    def wrapper(self):
        # Strip off the "test_" from the function name
        name = func.__name__[5:]

        def _die():
            salt.utils.process.appendproctitle("test_{}".format(name))

        attrname = "die_" + name
        setattr(self, attrname, _die)
        self.addCleanup(delattr, self, attrname)

    return wrapper


def incr(func):
    """
    Increment counter
    """

    @functools.wraps(func)
    def wrapper(self):
        # Strip off the "test_" from the function name
        name = func.__name__[5:]

        def _incr(counter, num):
            salt.utils.process.appendproctitle("test_{}".format(name))
            for _ in range(0, num):
                counter.value += 1

        attrname = "incr_" + name
        setattr(self, attrname, _incr)
        self.addCleanup(delattr, self, attrname)

    return wrapper


def spin(func):
    """
    Spin indefinitely
    """

    @functools.wraps(func)
    def wrapper(self):
        # Strip off the "test_" from the function name
        name = func.__name__[5:]

        def _spin():
            salt.utils.process.appendproctitle("test_{}".format(name))
            while True:
                time.sleep(1)

        attrname = "spin_" + name
        setattr(self, attrname, _spin)
        self.addCleanup(delattr, self, attrname)

    return wrapper


class TestProcessManager(TestCase):
    @spin
    @pytest.mark.slow_test
    def test_basic(self):
        """
        Make sure that the process is alive 2s later
        """
        process_manager = salt.utils.process.ProcessManager()
        self.addCleanup(process_manager.terminate)
        process_manager.add_process(self.spin_basic)
        initial_pid = next(iter(process_manager._process_map.keys()))
        time.sleep(2)
        process_manager.check_children()
        assert initial_pid == next(iter(process_manager._process_map.keys()))

    @spin
    def test_kill(self):
        process_manager = salt.utils.process.ProcessManager()
        self.addCleanup(process_manager.terminate)
        process_manager.add_process(self.spin_kill)
        initial_pid = next(iter(process_manager._process_map.keys()))
        # kill the child
        if salt.utils.platform.is_windows():
            os.kill(initial_pid, signal.SIGTERM)
        else:
            os.kill(initial_pid, signal.SIGKILL)
        # give the OS time to give the signal...
        time.sleep(0.1)
        process_manager.check_children()
        assert initial_pid != next(iter(process_manager._process_map.keys()))

    @die
    def test_restarting(self):
        """
        Make sure that the process is alive 2s later
        """
        process_manager = salt.utils.process.ProcessManager()
        self.addCleanup(process_manager.terminate)
        process_manager.add_process(self.die_restarting)
        initial_pid = next(iter(process_manager._process_map.keys()))
        time.sleep(2)
        process_manager.check_children()
        assert initial_pid != next(iter(process_manager._process_map.keys()))

    @incr
    def test_counter(self):
        counter = multiprocessing.Value("i", 0)
        process_manager = salt.utils.process.ProcessManager()
        self.addCleanup(process_manager.terminate)
        process_manager.add_process(self.incr_counter, args=(counter, 2))
        time.sleep(1)
        process_manager.check_children()
        time.sleep(1)
        # we should have had 2 processes go at it
        assert counter.value == 4


class TestThreadPool(TestCase):
    @pytest.mark.slow_test
    def test_basic(self):
        """
        Make sure the threadpool can do things
        """

        def incr_counter(counter):
            counter.value += 1

        counter = multiprocessing.Value("i", 0)

        pool = salt.utils.process.ThreadPool()
        sent = pool.fire_async(incr_counter, args=(counter,))
        self.assertTrue(sent)
        time.sleep(1)  # Sleep to let the threads do things
        self.assertEqual(counter.value, 1)
        self.assertEqual(pool._job_queue.qsize(), 0)

    @pytest.mark.slow_test
    def test_full_queue(self):
        """
        Make sure that a full threadpool acts as we expect
        """

        def incr_counter(counter):
            counter.value += 1

        counter = multiprocessing.Value("i", 0)

        # Create a pool with no workers and 1 queue size
        pool = salt.utils.process.ThreadPool(0, 1)
        # make sure we can put the one item in
        sent = pool.fire_async(incr_counter, args=(counter,))
        self.assertTrue(sent)
        # make sure we can't put more in
        sent = pool.fire_async(incr_counter, args=(counter,))
        self.assertFalse(sent)
        time.sleep(1)  # Sleep to let the threads do things
        # make sure no one updated the counter
        self.assertEqual(counter.value, 0)
        # make sure the queue is still full
        self.assertEqual(pool._job_queue.qsize(), 1)


class TestProcess(TestCase):
    def test_daemonize_if(self):
        # pylint: disable=assignment-from-none
        with patch("sys.argv", ["salt-call"]):
            ret = salt.utils.process.daemonize_if({})
            self.assertEqual(None, ret)

        ret = salt.utils.process.daemonize_if({"multiprocessing": False})
        self.assertEqual(None, ret)

        with patch("sys.platform", "win"):
            ret = salt.utils.process.daemonize_if({})
            self.assertEqual(None, ret)

        with patch("salt.utils.process.daemonize"), patch("sys.platform", "linux2"):
            salt.utils.process.daemonize_if({})
            self.assertTrue(salt.utils.process.daemonize.called)
        # pylint: enable=assignment-from-none


class TestProcessCallbacks(TestCase):
    @staticmethod
    def process_target(evt):
        evt.set()

    def test_callbacks(self):
        "Validate Process call after fork and finalize methods"
        teardown_to_mock = "salt.log.setup.shutdown_multiprocessing_logging"
        log_to_mock = "salt.log.setup.setup_multiprocessing_logging"
        with patch(teardown_to_mock) as ma, patch(log_to_mock) as mb:
            evt = multiprocessing.Event()
            proc = salt.utils.process.Process(target=self.process_target, args=(evt,))
            proc.run()
            assert evt.is_set()
        mb.assert_called()
        ma.assert_called()

    def test_callbacks_called_when_run_overridden(self):
        "Validate Process sub classes call after fork and finalize methods when run is overridden"

        class MyProcess(salt.utils.process.Process):
            def __init__(self):
                super().__init__()
                self.evt = multiprocessing.Event()

            def run(self):
                self.evt.set()

        teardown_to_mock = "salt.log.setup.shutdown_multiprocessing_logging"
        log_to_mock = "salt.log.setup.setup_multiprocessing_logging"
        with patch(teardown_to_mock) as ma, patch(log_to_mock) as mb:
            proc = MyProcess()
            proc.run()
            assert proc.evt.is_set()
        ma.assert_called()
        mb.assert_called()


@skipIf(not HAS_PSUTIL, "Missing psutil")
class TestSignalHandlingProcess(TestCase):
    @classmethod
    def Process(cls, pid):
        raise psutil.NoSuchProcess(pid)

    @classmethod
    def target(cls):
        os.kill(os.getpid(), signal.SIGTERM)

    @classmethod
    def children(cls, *args, **kwargs):
        raise psutil.NoSuchProcess(1)

    def test_process_does_not_exist(self):
        try:
            with patch("psutil.Process", self.Process):
                proc = salt.utils.process.SignalHandlingProcess(target=self.target)
                proc.start()
        except psutil.NoSuchProcess:
            assert False, "psutil.NoSuchProcess raised"

    def test_process_children_do_not_exist(self):
        try:
            with patch("psutil.Process.children", self.children):
                proc = salt.utils.process.SignalHandlingProcess(target=self.target)
                proc.start()
        except psutil.NoSuchProcess:
            assert False, "psutil.NoSuchProcess raised"

    @staticmethod
    def run_forever_sub_target(evt):
        "Used by run_forever_target to create a sub-process"
        while not evt.is_set():
            time.sleep(1)

    @staticmethod
    def run_forever_target(sub_target, evt):
        "A target that will run forever or until an event is set"
        p = multiprocessing.Process(target=sub_target, args=(evt,))
        p.start()
        p.join()

    @staticmethod
    def kill_target_sub_proc():
        pid = os.fork()
        if pid == 0:
            return
        pid = os.fork()
        if pid == 0:
            return
        time.sleep(0.1)
        try:
            os.kill(os.getpid(), signal.SIGINT)
        except KeyboardInterrupt:
            pass

    @skipIf(sys.platform.startswith("win"), "No os.fork on Windows")
    @pytest.mark.slow_test
    def test_signal_processing_regression_test(self):
        evt = multiprocessing.Event()
        sh_proc = salt.utils.process.SignalHandlingProcess(
            target=self.run_forever_target, args=(self.run_forever_sub_target, evt)
        )
        sh_proc.start()
        proc = multiprocessing.Process(target=self.kill_target_sub_proc)
        proc.start()
        proc.join()
        # When the bug exists, the kill_target_sub_proc signal will kill both
        # processes. sh_proc will be alive if the bug is fixed
        try:
            assert sh_proc.is_alive()
        finally:
            evt.set()
            sh_proc.join()

    @staticmethod
    def no_op_target():
        pass

    @staticmethod
    def pid_setting_target(sub_target, val, evt):
        val.value = os.getpid()
        p = multiprocessing.Process(target=sub_target, args=(evt,))
        p.start()
        p.join()

    @skipIf(sys.platform.startswith("win"), "Required signals not supported on windows")
    @pytest.mark.slow_test
    def test_signal_processing_handle_signals_called(self):
        "Validate SignalHandlingProcess handles signals"
        # Gloobal event to stop all processes we're creating
        evt = multiprocessing.Event()

        # Create a process to test signal handler
        val = multiprocessing.Value("i", 0)
        proc = salt.utils.process.SignalHandlingProcess(
            target=self.pid_setting_target,
            args=(self.run_forever_sub_target, val, evt),
        )
        proc.start()

        # Create a second process that should not respond to SIGINT or SIGTERM
        proc2 = multiprocessing.Process(
            target=self.run_forever_target,
            args=(self.run_forever_sub_target, evt),
        )
        proc2.start()

        # Wait for the sub process to set its pid
        while not val.value:
            time.sleep(0.3)

        assert not proc.signal_handled()

        # Send a signal that should get handled by the subprocess
        os.kill(val.value, signal.SIGTERM)

        # wait up to 10 seconds for signal handler:
        start = time.time()
        while time.time() - start < 10:
            if proc.signal_handled():
                break
            time.sleep(0.3)

        try:
            # Allow some time for the signal handler to do its thing
            assert proc.signal_handled()
            # Reap the signaled process
            proc.join(1)
            assert proc2.is_alive()
        finally:
            evt.set()
            proc2.join(30)
            proc.join(30)


class TestSignalHandlingProcessCallbacks(TestCase):
    @staticmethod
    def process_target(evt):
        evt.set()

    def test_callbacks(self):
        "Validate SignalHandlingProcess call after fork and finalize methods"

        teardown_to_mock = "salt.log.setup.shutdown_multiprocessing_logging"
        log_to_mock = "salt.log.setup.setup_multiprocessing_logging"
        sig_to_mock = "salt.utils.process.SignalHandlingProcess._setup_signals"
        # Mock _setup_signals so we do not register one for this process.
        evt = multiprocessing.Event()
        with patch(sig_to_mock):
            with patch(teardown_to_mock) as ma, patch(log_to_mock) as mb:
                sh_proc = salt.utils.process.SignalHandlingProcess(
                    target=self.process_target, args=(evt,)
                )
                sh_proc.run()
                assert evt.is_set()
        ma.assert_called()
        mb.assert_called()

    def test_callbacks_called_when_run_overridden(self):
        "Validate SignalHandlingProcess sub classes call after fork and finalize methods when run is overridden"

        class MyProcess(salt.utils.process.SignalHandlingProcess):
            def __init__(self):
                super().__init__()
                self.evt = multiprocessing.Event()

            def run(self):
                self.evt.set()

        teardown_to_mock = "salt.log.setup.shutdown_multiprocessing_logging"
        log_to_mock = "salt.log.setup.setup_multiprocessing_logging"
        sig_to_mock = "salt.utils.process.SignalHandlingProcess._setup_signals"
        # Mock _setup_signals so we do not register one for this process.
        with patch(sig_to_mock):
            with patch(teardown_to_mock) as ma, patch(log_to_mock) as mb:
                sh_proc = MyProcess()
                sh_proc.run()
                assert sh_proc.evt.is_set()
        ma.assert_called()
        mb.assert_called()


class TestDup2(TestCase):
    def test_dup2_no_fileno(self):
        "The dup2 method does not fail on streams without fileno support"
        f1 = io.StringIO("some initial text data")
        f2 = io.StringIO("some initial other text data")
        with self.assertRaises(io.UnsupportedOperation):
            f1.fileno()
        with patch("os.dup2") as dup_mock:
            try:
                salt.utils.process.dup2(f1, f2)
            except io.UnsupportedOperation:
                assert False, "io.UnsupportedOperation was raised"
        assert not dup_mock.called


def null_target():
    pass


def event_target(event):
    while True:
        if event.wait(5):
            break


class TestProcessList(TestCase):
    @staticmethod
    def wait_for_proc(proc, timeout=10):
        start = time.time()
        while proc.is_alive():
            if time.time() - start > timeout:
                raise Exception("Process did not finishe before timeout")
            time.sleep(0.3)

    @pytest.mark.slow_test
    def test_process_list_process(self):
        plist = salt.utils.process.SubprocessList()
        proc = multiprocessing.Process(target=null_target)
        proc.start()
        plist.add(proc)
        assert proc in plist.processes
        self.wait_for_proc(proc)
        assert not proc.is_alive()
        plist.cleanup()
        assert proc not in plist.processes

    def test_process_list_thread(self):
        plist = salt.utils.process.SubprocessList()
        thread = threading.Thread(target=null_target)
        thread.start()
        plist.add(thread)
        assert thread in plist.processes
        self.wait_for_proc(thread)
        assert not thread.is_alive()
        plist.cleanup()
        assert thread not in plist.processes

    @pytest.mark.slow_test
    def test_process_list_cleanup(self):
        plist = salt.utils.process.SubprocessList()
        event = multiprocessing.Event()
        proc = multiprocessing.Process(target=event_target, args=[event])
        proc.start()
        plist.add(proc)
        assert proc in plist.processes
        plist.cleanup()
        event.set()
        assert proc in plist.processes
        self.wait_for_proc(proc)
        assert not proc.is_alive()
        plist.cleanup()
        assert proc not in plist.processes


class TestDeprecatedClassNames(TestCase):
    @staticmethod
    def process_target():
        pass

    @staticmethod
    def patched_warn_until_date(current_date):
        def _patched_warn_until_date(
            date,
            message,
            category=DeprecationWarning,
            stacklevel=None,
            _current_date=current_date,
            _dont_call_warnings=False,
        ):
            # Because we add another function in between, the stacklevel
            # set in salt.utils.process, 3, needs to now be 4
            stacklevel = 4
            return warn_until_date(
                date,
                message,
                category=category,
                stacklevel=stacklevel,
                _current_date=_current_date,
                _dont_call_warnings=_dont_call_warnings,
            )

        return _patched_warn_until_date

    def test_multiprocessing_process_warning(self):
        # We *always* want *all* warnings thrown on this module
        warnings.filterwarnings("always", "", DeprecationWarning, __name__)

        fake_utcnow = datetime.date(2021, 1, 1)

        proc = None

        try:
            with patch(
                "salt.utils.versions.warn_until_date",
                self.patched_warn_until_date(fake_utcnow),
            ):
                # Test warning
                with warnings.catch_warnings(record=True) as recorded_warnings:
                    proc = salt.utils.process.MultiprocessingProcess(
                        target=self.process_target
                    )
                    self.assertEqual(
                        "Please stop using 'salt.utils.process.MultiprocessingProcess' "
                        "and instead use 'salt.utils.process.Process'. "
                        "'salt.utils.process.MultiprocessingProcess' will go away "
                        "after 2022-01-01.",
                        str(recorded_warnings[0].message),
                    )
        finally:
            if proc is not None:
                del proc

    def test_multiprocessing_process_runtime_error(self):
        fake_utcnow = datetime.date(2022, 1, 1)

        proc = None

        try:
            with patch(
                "salt.utils.versions.warn_until_date",
                self.patched_warn_until_date(fake_utcnow),
            ):
                with self.assertRaisesRegex(
                    RuntimeError,
                    r"Please stop using 'salt.utils.process.MultiprocessingProcess' "
                    r"and instead use 'salt.utils.process.Process'. "
                    r"'salt.utils.process.MultiprocessingProcess' will go away "
                    r"after 2022-01-01. "
                    r"This warning\(now exception\) triggered on "
                    r"filename '(.*)test_process.py', line number ([\d]+), is "
                    r"supposed to be shown until ([\d-]+). Today is ([\d-]+). "
                    r"Please remove the warning.",
                ):
                    proc = salt.utils.process.MultiprocessingProcess(
                        target=self.process_target
                    )
        finally:
            if proc is not None:
                del proc

    def test_signal_handling_multiprocessing_process_warning(self):
        # We *always* want *all* warnings thrown on this module
        warnings.filterwarnings("always", "", DeprecationWarning, __name__)

        fake_utcnow = datetime.date(2021, 1, 1)

        proc = None

        try:
            with patch(
                "salt.utils.versions.warn_until_date",
                self.patched_warn_until_date(fake_utcnow),
            ):
                # Test warning
                with warnings.catch_warnings(record=True) as recorded_warnings:
                    proc = salt.utils.process.SignalHandlingMultiprocessingProcess(
                        target=self.process_target
                    )
                    self.assertEqual(
                        "Please stop using"
                        " 'salt.utils.process.SignalHandlingMultiprocessingProcess' and"
                        " instead use 'salt.utils.process.SignalHandlingProcess'."
                        " 'salt.utils.process.SignalHandlingMultiprocessingProcess'"
                        " will go away after 2022-01-01.",
                        str(recorded_warnings[0].message),
                    )
        finally:
            if proc is not None:
                del proc

    def test_signal_handling_multiprocessing_process_runtime_error(self):
        fake_utcnow = datetime.date(2022, 1, 1)

        proc = None

        try:
            with patch(
                "salt.utils.versions.warn_until_date",
                self.patched_warn_until_date(fake_utcnow),
            ):
                with self.assertRaisesRegex(
                    RuntimeError,
                    r"Please stop using"
                    r" 'salt.utils.process.SignalHandlingMultiprocessingProcess' "
                    r"and instead use 'salt.utils.process.SignalHandlingProcess'. "
                    r"'salt.utils.process.SignalHandlingMultiprocessingProcess' will go"
                    r" away "
                    r"after 2022-01-01. "
                    r"This warning\(now exception\) triggered on "
                    r"filename '(.*)test_process.py', line number ([\d]+), is "
                    r"supposed to be shown until ([\d-]+). Today is ([\d-]+). "
                    r"Please remove the warning.",
                ):
                    proc = salt.utils.process.SignalHandlingMultiprocessingProcess(
                        target=self.process_target
                    )
        finally:
            if proc is not None:
                del proc


class CMORProcessHelper:
    def __init__(self, file_name):
        self._lock = threading.Lock()
        self._running = True
        self._queue = multiprocessing.Queue()
        self._ret_queue = multiprocessing.Queue()
        self._process = multiprocessing.Process(
            target=self.test_process,
            args=(file_name, self._queue, self._ret_queue),
            daemon=True,
        )
        self._process.start()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def claim(self):
        try:
            self._lock.acquire()
            if self._running:
                self._queue.put("claim")
                return self._ret_queue.get(timeout=10)
        finally:
            self._lock.release()

    def stop(self):
        try:
            self._lock.acquire()
            if self._running:
                self._running = False

                self._queue.put("stop")
                self._process.join(timeout=10)

                self._queue.close()
                self._ret_queue.close()
        finally:
            self._lock.release()

    @property
    def pid(self):
        return self._process.pid

    @staticmethod
    def test_process(file_name, queue, ret_queue):
        while True:
            action = queue.get()
            if action == "claim":
                ret_queue.put(
                    salt.utils.process.claim_mantle_of_responsibility(file_name)
                )
            elif action == "stop":
                return


@skipIf(not HAS_PSUTIL, "Missing psutil")
class TestGetProcessInfo(TestCase):
    def setUp(self):
        handle, self.cmor_test_file_path = tempfile.mkstemp()
        os.close(handle)
        self.addCleanup(os.unlink, self.cmor_test_file_path)

    def test_this_process(self):
        this_process_info = salt.utils.process.get_process_info()

        self.assertEqual(
            this_process_info, salt.utils.process.get_process_info(os.getpid())
        )
        self.assertIsNotNone(this_process_info)

        for key in ("pid", "name", "start_time"):
            self.assertIn(key, this_process_info)

        raw_process_info = psutil.Process(os.getpid())
        self.assertEqual(this_process_info["pid"], os.getpid())
        self.assertEqual(this_process_info["name"], raw_process_info.name())
        self.assertEqual(
            this_process_info["start_time"], raw_process_info.create_time()
        )

    def test_random_processes(self):
        for _ in range(3):
            with CMORProcessHelper(self.cmor_test_file_path) as p1:
                pid = p1.pid
                self.assertIsInstance(salt.utils.process.get_process_info(pid), dict)
            self.assertIsNone(salt.utils.process.get_process_info(pid))


@skipIf(not HAS_PSUTIL, "Missing psutil")
class TestClaimMantleOfResponsibility(TestCase):
    def setUp(self):
        handle, self.cmor_test_file_path = tempfile.mkstemp()
        os.close(handle)
        self.addCleanup(os.unlink, self.cmor_test_file_path)

    def test_simple_claim_no_psutil(self):
        salt.utils.process.claim_mantle_of_responsibility(self.cmor_test_file_path)

    def test_simple_claim(self):
        for _ in range(5):
            self.assertTrue(
                salt.utils.process.claim_mantle_of_responsibility(
                    self.cmor_test_file_path
                )
            )

    def test_multiple_processes(self):
        with CMORProcessHelper(self.cmor_test_file_path) as p1:
            self.assertTrue(p1.claim())
            self.assertFalse(
                salt.utils.process.claim_mantle_of_responsibility(
                    self.cmor_test_file_path
                )
            )
            with CMORProcessHelper(self.cmor_test_file_path) as p2:
                for _ in range(3):
                    self.assertFalse(p2.claim())
            self.assertTrue(p1.claim())

        with CMORProcessHelper(self.cmor_test_file_path) as p1:
            self.assertTrue(p1.claim())
            self.assertFalse(
                salt.utils.process.claim_mantle_of_responsibility(
                    self.cmor_test_file_path
                )
            )

        self.assertTrue(
            salt.utils.process.claim_mantle_of_responsibility(self.cmor_test_file_path)
        )


@skipIf(not HAS_PSUTIL, "Missing psutil")
class TestCheckMantleOfResponsibility(TestCase):
    def setUp(self):
        handle, self.cmor_test_file_path = tempfile.mkstemp()
        os.close(handle)
        self.addCleanup(os.unlink, self.cmor_test_file_path)

    def test_simple_claim_no_psutil(self):
        self.assertIsNone(
            salt.utils.process.check_mantle_of_responsibility(self.cmor_test_file_path)
        )

    def test_simple_claim(self):
        self.assertIsNone(
            salt.utils.process.check_mantle_of_responsibility(self.cmor_test_file_path)
        )
        salt.utils.process.claim_mantle_of_responsibility(self.cmor_test_file_path)
        pid = salt.utils.process.get_process_info()["pid"]
        self.assertEqual(
            pid,
            salt.utils.process.check_mantle_of_responsibility(self.cmor_test_file_path),
        )

    def test_multiple_processes(self):
        self.assertIsNone(
            salt.utils.process.check_mantle_of_responsibility(self.cmor_test_file_path)
        )

        with CMORProcessHelper(self.cmor_test_file_path) as p1:
            self.assertTrue(p1.claim())
            random_pid = salt.utils.process.check_mantle_of_responsibility(
                self.cmor_test_file_path
            )

            self.assertIsInstance(random_pid, int)

            with CMORProcessHelper(self.cmor_test_file_path) as p2:
                for _ in range(3):
                    self.assertFalse(p2.claim())
                self.assertEqual(
                    random_pid,
                    salt.utils.process.check_mantle_of_responsibility(
                        self.cmor_test_file_path
                    ),
                )

        self.assertIsNone(
            salt.utils.process.check_mantle_of_responsibility(self.cmor_test_file_path)
        )
        salt.utils.process.claim_mantle_of_responsibility(self.cmor_test_file_path)
        pid = salt.utils.process.get_process_info()["pid"]
        self.assertEqual(
            pid,
            salt.utils.process.check_mantle_of_responsibility(self.cmor_test_file_path),
        )



"""
Tests for the Mandrill execution module.
"""


import salt.modules.mandrill as mandrill
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

# Test data
TEST_SEND = {
    "result": True,
    "comment": "",
    "out": [
        {
            "status": "sent",
            "_id": "c4353540a3c123eca112bbdd704ab6",
            "email": "recv@example.com",
            "reject_reason": None,
        }
    ],
}


class MandrillModuleTest(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.mandrill.
    """

    def setup_loader_modules(self):
        module_globals = {
            mandrill: {
                "__salt__": {
                    "config.merge": MagicMock(
                        return_value={"mandrill": {"key": "2orgk34kgk34g"}}
                    )
                }
            }
        }
        if mandrill.HAS_REQUESTS is False:
            module_globals["sys.modules"] = {"requests": MagicMock()}
        return module_globals

    def test_send(self):
        """
        Test the send function.
        """
        mock_cmd = MagicMock(return_value=TEST_SEND)
        with patch.object(mandrill, "send", mock_cmd) as send:
            self.assertEqual(
                send(
                    message={
                        "subject": "Hi",
                        "from_email": "test@example.com",
                        "to": [{"email": "recv@example.com", "type": "to"}],
                    }
                ),
                TEST_SEND,
            )

    def test_deprecation_58640(self):
        # check that type error will be raised
        message = {
            "subject": "Hi",
            "from_email": "test@example.com",
            "to": [{"email": "recv@example.com", "type": "to"}],
        }
        self.assertRaises(
            TypeError, mandrill.send, **{"message": message, "async": True}
        )

        # check that async will raise an error
        try:
            mandrill.send(  # pylint: disable=unexpected-keyword-arg
                **{"message": message, "async": True}
            )
        except TypeError as no_async:
            self.assertEqual(
                str(no_async),
                "send() got an unexpected keyword argument 'async'",
            )



"""
Support for DEB packages
"""

import datetime
import logging
import os
import re

import salt.utils.args
import salt.utils.data
import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "lowpkg"


def __virtual__():
    """
    Confirm this module is on a Debian based system
    """
    if __grains__["os_family"] == "Debian":
        return __virtualname__
    return (
        False,
        "The dpkg execution module cannot be loaded: "
        "only works on Debian family systems.",
    )


def bin_pkg_info(path, saltenv="base"):
    """
    .. versionadded:: 2015.8.0

    Parses RPM metadata and returns a dictionary of information about the
    package (name, version, etc.).

    path
        Path to the file. Can either be an absolute path to a file on the
        minion, or a salt fileserver URL (e.g. ``salt://path/to/file.rpm``).
        If a salt fileserver URL is passed, the file will be cached to the
        minion so that it can be examined.

    saltenv : base
        Salt fileserver environment from which to retrieve the package. Ignored
        if ``path`` is a local file path on the minion.

    CLI Example:

    .. code-block:: bash

        salt '*' lowpkg.bin_pkg_info /root/foo-1.2.3-1ubuntu1_all.deb
        salt '*' lowpkg.bin_pkg_info salt://foo-1.2.3-1ubuntu1_all.deb
    """
    # If the path is a valid protocol, pull it down using cp.cache_file
    if __salt__["config.valid_fileproto"](path):
        newpath = __salt__["cp.cache_file"](path, saltenv)
        if not newpath:
            raise CommandExecutionError(
                "Unable to retrieve {} from saltenv '{}'".format(path, saltenv)
            )
        path = newpath
    else:
        if not os.path.exists(path):
            raise CommandExecutionError("{} does not exist on minion".format(path))
        elif not os.path.isabs(path):
            raise SaltInvocationError("{} does not exist on minion".format(path))

    cmd = ["dpkg", "-I", path]
    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace")
    if result["retcode"] != 0:
        msg = "Unable to get info for " + path
        if result["stderr"]:
            msg += ": " + result["stderr"]
        raise CommandExecutionError(msg)

    ret = {}
    for line in result["stdout"].splitlines():
        line = line.strip()
        if re.match(r"^Package[ ]*:", line):
            ret["name"] = line.split()[-1]
        elif re.match(r"^Version[ ]*:", line):
            ret["version"] = line.split()[-1]
        elif re.match(r"^Architecture[ ]*:", line):
            ret["arch"] = line.split()[-1]

    missing = [x for x in ("name", "version", "arch") if x not in ret]
    if missing:
        raise CommandExecutionError(
            "Unable to get {} for {}".format(", ".join(missing), path)
        )

    if __grains__.get("cpuarch", "") == "x86_64":
        osarch = __grains__.get("osarch", "")
        arch = ret["arch"]
        if arch != "all" and osarch == "amd64" and osarch != arch:
            ret["name"] += ":{}".format(arch)

    return ret


def unpurge(*packages):
    """
    Change package selection for each package specified to 'install'

    CLI Example:

    .. code-block:: bash

        salt '*' lowpkg.unpurge curl
    """
    if not packages:
        return {}
    old = __salt__["pkg.list_pkgs"](purge_desired=True)
    ret = {}
    __salt__["cmd.run"](
        ["dpkg", "--set-selections"],
        stdin=r"\n".join(["{} install".format(x) for x in packages]),
        python_shell=False,
        output_loglevel="trace",
    )
    __context__.pop("pkg.list_pkgs", None)
    new = __salt__["pkg.list_pkgs"](purge_desired=True)
    return salt.utils.data.compare_dicts(old, new)


def list_pkgs(*packages, **kwargs):
    """
    List the packages currently installed in a dict::

        {'<package_name>': '<version>'}

    External dependencies::

        Virtual package resolution requires aptitude. Because this function
        uses dpkg, virtual packages will be reported as not installed.

    CLI Example:

    .. code-block:: bash

        salt '*' lowpkg.list_pkgs
        salt '*' lowpkg.list_pkgs hostname
        salt '*' lowpkg.list_pkgs hostname mount
    """
    cmd = [
        "dpkg-query",
        "-f=${db:Status-Status}\t${binary:Package}\t${Version}\n",
        "-W",
    ] + list(packages)
    out = __salt__["cmd.run_all"](cmd, python_shell=False)
    if out["retcode"] != 0:
        msg = "Error:  " + out["stderr"]
        log.error(msg)
        return msg

    lines = [line.split("\t", 1) for line in out["stdout"].splitlines()]
    pkgs = dict([line.split("\t") for status, line in lines if status == "installed"])

    return pkgs


def file_list(*packages, **kwargs):
    """
    List the files that belong to a package. Not specifying any packages will
    return a list of _every_ file on the system's package database (not
    generally recommended).

    CLI Examples:

    .. code-block:: bash

        salt '*' lowpkg.file_list hostname
        salt '*' lowpkg.file_list hostname mount
        salt '*' lowpkg.file_list
    """
    errors = []
    ret = set()
    cmd = ["dpkg-query", "-f=${db:Status-Status}\t${binary:Package}\n", "-W"] + list(
        packages
    )
    out = __salt__["cmd.run_all"](cmd, python_shell=False)
    if out["retcode"] != 0:
        msg = "Error:  " + out["stderr"]
        log.error(msg)
        return msg

    lines = [line.split("\t") for line in out["stdout"].splitlines()]
    pkgs = [package for (status, package) in lines if status == "installed"]

    for pkg in pkgs:
        output = __salt__["cmd.run"](["dpkg", "-L", pkg], python_shell=False)
        fileset = set(output.splitlines())
        ret = ret.union(fileset)
    return {"errors": errors, "files": sorted(ret)}


def file_dict(*packages, **kwargs):
    """
    List the files that belong to a package, grouped by package. Not
    specifying any packages will return a list of _every_ file on the system's
    package database (not generally recommended).

    CLI Examples:

    .. code-block:: bash

        salt '*' lowpkg.file_dict hostname
        salt '*' lowpkg.file_dict hostname mount
        salt '*' lowpkg.file_dict
    """
    errors = []
    ret = {}
    cmd = ["dpkg-query", "-f=${db:Status-Status}\t${binary:Package}\n", "-W"] + list(
        packages
    )
    out = __salt__["cmd.run_all"](cmd, python_shell=False)
    if out["retcode"] != 0:
        msg = "Error:  " + out["stderr"]
        log.error(msg)
        return msg

    lines = [line.split("\t") for line in out["stdout"].splitlines()]
    pkgs = [package for (status, package) in lines if status == "installed"]

    for pkg in pkgs:
        cmd = ["dpkg", "-L", pkg]
        ret[pkg] = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    return {"errors": errors, "packages": ret}


def _get_pkg_info(*packages, **kwargs):
    """
    Return list of package information. If 'packages' parameter is empty,
    then data about all installed packages will be returned.

    :param packages: Specified packages.
    :param failhard: Throw an exception if no packages found.
    :return:
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    failhard = kwargs.pop("failhard", True)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    if __grains__["os"] == "Ubuntu" and __grains__["osrelease_info"] < (12, 4):
        bin_var = "${binary}"
    else:
        bin_var = "${Package}"

    ret = []
    cmd = (
        "dpkg-query -W -f='package:" + bin_var + "\\n"
        "revision:${binary:Revision}\\n"
        "architecture:${Architecture}\\n"
        "maintainer:${Maintainer}\\n"
        "summary:${Summary}\\n"
        "source:${source:Package}\\n"
        "version:${Version}\\n"
        "section:${Section}\\n"
        "installed_size:${Installed-size}\\n"
        "size:${Size}\\n"
        "MD5:${MD5sum}\\n"
        "SHA1:${SHA1}\\n"
        "SHA256:${SHA256}\\n"
        "origin:${Origin}\\n"
        "homepage:${Homepage}\\n"
        "status:${db:Status-Abbrev}\\n"
        "======\\n"
        "description:${Description}\\n"
        "------\\n'"
    )
    cmd += " {}".format(" ".join(packages))
    cmd = cmd.strip()

    call = __salt__["cmd.run_all"](cmd, python_chell=False)
    if call["retcode"]:
        if failhard:
            raise CommandExecutionError(
                "Error getting packages information: {}".format(call["stderr"])
            )
        else:
            return ret

    for pkg_info in [elm for elm in re.split(r"------", call["stdout"]) if elm.strip()]:
        pkg_data = {}
        pkg_info, pkg_descr = re.split(r"======", pkg_info)
        for pkg_info_line in [
            el.strip() for el in pkg_info.split(os.linesep) if el.strip()
        ]:
            key, value = pkg_info_line.split(":", 1)
            if value:
                pkg_data[key] = value
            install_date = _get_pkg_install_time(pkg_data.get("package"))
            if install_date:
                pkg_data["install_date"] = install_date
        pkg_data["description"] = pkg_descr.split(":", 1)[-1]
        ret.append(pkg_data)

    return ret


def _get_pkg_license(pkg):
    """
    Try to get a license from the package.
    Based on https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

    :param pkg:
    :return:
    """
    licenses = set()
    cpr = "/usr/share/doc/{}/copyright".format(pkg)
    if os.path.exists(cpr):
        with salt.utils.files.fopen(cpr) as fp_:
            for line in salt.utils.stringutils.to_unicode(fp_.read()).split(os.linesep):
                if line.startswith("License:"):
                    licenses.add(line.split(":", 1)[1].strip())

    return ", ".join(sorted(licenses))


def _get_pkg_install_time(pkg):
    """
    Return package install time, based on the /var/lib/dpkg/info/<package>.list

    :return:
    """
    iso_time = None
    if pkg is not None:
        location = "/var/lib/dpkg/info/{}.list".format(pkg)
        if os.path.exists(location):
            iso_time = (
                datetime.datetime.utcfromtimestamp(
                    int(os.path.getmtime(location))
                ).isoformat()
                + "Z"
            )

    return iso_time


def _get_pkg_ds_avail():
    """
    Get the package information of the available packages, maintained by dselect.
    Note, this will be not very useful, if dselect isn't installed.

    :return:
    """
    avail = "/var/lib/dpkg/available"
    if not salt.utils.path.which("dselect") or not os.path.exists(avail):
        return dict()

    # Do not update with dselect, just read what is.
    ret = dict()
    pkg_mrk = "Package:"
    pkg_name = "package"
    with salt.utils.files.fopen(avail) as fp_:
        for pkg_info in salt.utils.stringutils.to_unicode(fp_.read()).split(pkg_mrk):
            nfo = dict()
            for line in (pkg_mrk + pkg_info).split(os.linesep):
                line = line.split(": ", 1)
                if len(line) != 2:
                    continue
                key, value = line
                if value.strip():
                    nfo[key.lower()] = value
            if nfo.get(pkg_name):
                ret[nfo[pkg_name]] = nfo

    return ret


def info(*packages, **kwargs):
    """
    Returns a detailed summary of package information for provided package names.
    If no packages are specified, all packages will be returned.

    .. versionadded:: 2015.8.1

    packages
        The names of the packages for which to return information.

    failhard
        Whether to throw an exception if none of the packages are installed.
        Defaults to True.

        .. versionadded:: 2016.11.3

    CLI Example:

    .. code-block:: bash

        salt '*' lowpkg.info
        salt '*' lowpkg.info apache2 bash
        salt '*' lowpkg.info 'php5*' failhard=false
    """
    # Get the missing information from the /var/lib/dpkg/available, if it is there.
    # However, this file is operated by dselect which has to be installed.
    dselect_pkg_avail = _get_pkg_ds_avail()

    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    failhard = kwargs.pop("failhard", True)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)

    ret = dict()
    for pkg in _get_pkg_info(*packages, failhard=failhard):
        # Merge extra information from the dselect, if available
        for pkg_ext_k, pkg_ext_v in dselect_pkg_avail.get(pkg["package"], {}).items():
            if pkg_ext_k not in pkg:
                pkg[pkg_ext_k] = pkg_ext_v
        # Remove "technical" keys
        for t_key in [
            "installed_size",
            "depends",
            "recommends",
            "provides",
            "replaces",
            "conflicts",
            "bugs",
            "description-md5",
            "task",
        ]:
            if t_key in pkg:
                del pkg[t_key]

        lic = _get_pkg_license(pkg["package"])
        if lic:
            pkg["license"] = lic
        ret[pkg["package"]] = pkg

    return ret



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""


import salt.modules.pagerduty as pagerduty
import salt.utils.json
import salt.utils.pagerduty
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PagerdutyTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.pagerduty
    """

    def setup_loader_modules(self):
        return {
            pagerduty: {"__salt__": {"config.option": MagicMock(return_value=None)}}
        }

    def test_list_services(self):
        """
        Test for List services belonging to this account
        """
        with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
            self.assertEqual(pagerduty.list_services(), "A")

    def test_list_incidents(self):
        """
        Test for List incidents belonging to this account
        """
        with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
            self.assertEqual(pagerduty.list_incidents(), "A")

    def test_list_users(self):
        """
        Test for List users belonging to this account
        """
        with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
            self.assertEqual(pagerduty.list_users(), "A")

    def test_list_schedules(self):
        """
        Test for List schedules belonging to this account
        """
        with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
            self.assertEqual(pagerduty.list_schedules(), "A")

    def test_list_windows(self):
        """
        Test for List maintenance windows belonging to this account
        """
        with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
            self.assertEqual(pagerduty.list_windows(), "A")

    def test_list_policies(self):
        """
        Test for List escalation policies belonging to this account
        """
        with patch.object(salt.utils.pagerduty, "list_items", return_value="A"):
            self.assertEqual(pagerduty.list_policies(), "A")

    def test_create_event(self):
        """
        Test for Create an event in PagerDuty. Designed for use in states.
        """
        with patch.object(salt.utils.json, "loads", return_value=["A"]):
            with patch.object(salt.utils.pagerduty, "query", return_value="A"):
                self.assertListEqual(pagerduty.create_event(), ["A"])



"""
    :codeauthor: `Tyler Johnson <tjohnson@saltstack.com>`

    tests.unit.cloud.clouds.openstack_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


from salt.cloud.clouds import openstack
from salt.utils import dictupdate
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

# pylint: disable=confusing-with-statement


class MockImage:
    name = "image name"
    id = "image id"


class MockNode:
    name = "node name"
    id = "node id"
    flavor = MockImage()
    status = "node status"

    def __init__(self, image):
        self.image = image

    def __iter__(self):
        return iter(())


class MockConn:
    def __init__(self, image):
        self.node = MockNode(image)

    def get_image(self, *args, **kwargs):
        return self.node.image

    def get_flavor(self, *args, **kwargs):
        return self.node.flavor

    def get_server(self, *args, **kwargs):
        return self.node

    def list_servers(self, *args, **kwargs):
        return [self.node]


class OpenstackTestCase(TestCase, LoaderModuleMockMixin):
    """
    Unit TestCase for salt.cloud.clouds.openstack module.
    """

    def setup_loader_modules(self):
        return {
            openstack: {
                "__active_provider_name__": "",
                "__opts__": {
                    "providers": {
                        "my-openstack-cloud": {
                            "openstack": {
                                "auth": "daenerys",
                                "region_name": "westeros",
                                "cloud": "openstack",
                            }
                        }
                    }
                },
            }
        }

    def test_get_configured_provider_bad(self):
        with patch.dict(openstack.__opts__, {"providers": {}}):
            result = openstack.get_configured_provider()
            self.assertEqual(result, False)

    def test_get_configured_provider_auth(self):
        config = {
            "region_name": "westeros",
            "auth": "daenerys",
        }
        with patch.dict(
            openstack.__opts__,
            {"providers": {"my-openstack-cloud": {"openstack": config}}},
        ):
            result = openstack.get_configured_provider()
            self.assertEqual(config, result)

    def test_get_configured_provider_cloud(self):
        config = {
            "region_name": "westeros",
            "cloud": "foo",
        }
        with patch.dict(
            openstack.__opts__,
            {"providers": {"my-openstack-cloud": {"openstack": config}}},
        ):
            result = openstack.get_configured_provider()
            self.assertEqual(config, result)

    def test_get_dependencies(self):
        HAS_SHADE = (True, "Please install newer version of shade: >= 1.19.0")
        with patch("salt.cloud.clouds.openstack.HAS_SHADE", HAS_SHADE):
            result = openstack.get_dependencies()
            self.assertEqual(result, True)

    def test_get_dependencies_no_shade(self):
        HAS_SHADE = (False, "Install pypi module shade >= 1.19.0")
        with patch("salt.cloud.clouds.openstack.HAS_SHADE", HAS_SHADE):
            result = openstack.get_dependencies()
            self.assertEqual(result, False)

    def test_list_nodes_full_image_str(self):
        node_image = "node image"
        conn = MockConn(node_image)
        with patch("salt.cloud.clouds.openstack._get_ips", return_value=[]):
            ret = openstack.list_nodes_full(conn=conn)
            self.assertEqual(ret[conn.node.name]["image"], node_image)

    def test_list_nodes_full_image_obj(self):
        conn = MockConn(MockImage())
        with patch("salt.cloud.clouds.openstack._get_ips", return_value=[]):
            ret = openstack.list_nodes_full(conn=conn)
            self.assertEqual(ret[conn.node.name]["image"], MockImage.name)

    def test_show_instance(self):
        conn = MockConn(MockImage())
        with patch("salt.cloud.clouds.openstack._get_ips", return_value=[]):
            ret = openstack.show_instance(conn.node.name, conn=conn, call="action")
            self.assertEqual(ret["image"], MockImage.name)

    def test_request_instance_should_use_provided_connection_if_not_None(self):
        fake_conn = MagicMock()

        patch_get_conn = patch("salt.cloud.clouds.openstack.get_conn", autospec=True)
        patch_utils = patch.dict(
            openstack.__utils__,
            {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
        )
        patch_shade = patch.object(
            openstack, "shade.exc.OpenStackCloudException", Exception, create=True
        )

        with patch_get_conn as fake_get_conn, patch_utils, patch_shade:
            openstack.request_instance(
                vm_={"name": "fnord", "driver": "fnord"}, conn=fake_conn
            )

            fake_get_conn.assert_not_called()

    def test_request_instance_should_create_conn_if_provided_is_None(self):
        none_conn = None

        patch_get_conn = patch("salt.cloud.clouds.openstack.get_conn", autospec=True)
        patch_utils = patch.dict(
            openstack.__utils__,
            {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
        )
        patch_shade = patch.object(
            openstack, "shade.exc.OpenStackCloudException", Exception, create=True
        )

        with patch_get_conn as fake_get_conn, patch_utils, patch_shade:
            openstack.request_instance(
                vm_={"name": "fnord", "driver": "fnord"}, conn=none_conn
            )

            fake_get_conn.assert_called_once_with()

    # According to
    # https://docs.openstack.org/shade/latest/user/usage.html#shade.OpenStackCloud.create_server
    # the `network` parameter can be:
    # (optional) Network dict or name or ID to attach the server to.
    # Mutually exclusive with the nics parameter. Can also be be a list of
    # network names or IDs or network dicts.
    #
    # Here we're testing a normal dictionary
    def test_request_instance_should_be_able_to_provide_a_dictionary_for_network(self):
        fake_conn = MagicMock()
        expected_network = {"foo": "bar"}
        vm_ = {"name": "fnord", "driver": "fnord", "network": expected_network}
        patch_utils = patch.dict(
            openstack.__utils__,
            {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
        )
        with patch_utils:
            openstack.request_instance(vm_=vm_, conn=fake_conn)

            call_kwargs = fake_conn.create_server.mock_calls[0][-1]
            self.assertDictEqual(call_kwargs["network"], expected_network)

    # Here we're testing the list of dictionaries
    def test_request_instance_should_be_able_to_provide_a_list_of_dictionaries_for_network(
        self,
    ):
        fake_conn = MagicMock()
        expected_network = [{"foo": "bar"}, {"bang": "quux"}]
        vm_ = {"name": "fnord", "driver": "fnord", "network": expected_network}
        patch_utils = patch.dict(
            openstack.__utils__,
            {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
        )
        with patch_utils:
            openstack.request_instance(vm_=vm_, conn=fake_conn)

            call_kwargs = fake_conn.create_server.mock_calls[0][-1]
            assert call_kwargs["network"] == expected_network

    # Here we're testing for names/IDs
    def test_request_instance_should_be_able_to_provide_a_list_of_single_ids_or_names_for_network(
        self,
    ):
        fake_conn = MagicMock()
        expected_network = ["foo", "bar", "bang", "fnord1", "fnord2"]
        vm_ = {"name": "fnord", "driver": "fnord", "network": expected_network}
        patch_utils = patch.dict(
            openstack.__utils__,
            {"cloud.check_name": MagicMock(), "dictupdate.update": dictupdate.update},
        )
        with patch_utils:
            openstack.request_instance(vm_=vm_, conn=fake_conn)

            call_kwargs = fake_conn.create_server.mock_calls[0][-1]
            assert call_kwargs["network"] == expected_network

    # Testing that we get a dict that we expect for create_server
    def test__clean_create_kwargs(self):
        params = {
            "name": "elmer",
            "image": "mirrormirror",
            "flavor": "chocolate",
            "auto_ip": True,
            "ips": ["hihicats"],
            "ip_pool": "olympic",
            "root_volume": "iamgroot",
            "boot_volume": "pussnboots",
            "terminate_volume": False,
            "volumes": ["lots", "of", "books"],
            "meta": {"full": "meta"},
            "files": {"shred": "this"},
            "reservation_id": "licenseandregistration",
            "security_groups": ["wanna", "play", "repeat"],
            "key_name": "clortho",
            "availability_zone": "callmemaybe",
            "block_device_mapping": [{"listof": "dicts"}],
            "block_device_mapping_v2": [{"listof": "dicts"}],
            "nics": ["thats", "me"],
            "scheduler_hints": {"so": "many"},
            "config_drive": True,
            "disk_config": "donkey",
            "admin_pass": "password",
            "wait": False,
            "timeout": 30,
            "reuse_ips": True,
            "network": ["also", "a", "dict"],
            "boot_from_volume": True,
            "volume_size": 30,
            "nat_destination": "albuquerque",
            "group": "ledzeppelin",
            "userdata": "needmoreinput",
            "thisgetsdropped": "yup",
        }
        patch_utils = patch.dict(
            openstack.__utils__,
            {"dictupdate.update": dictupdate.update},
        )
        with patch_utils:
            ret = openstack._clean_create_kwargs(**params)
            params.pop("thisgetsdropped")
            self.assertDictEqual(params, ret)



"""
tests.pytests.unit.test_config
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for salt's config modulet
"""
import salt.config


def test_call_id_function(tmp_path):
    "Defining id_function works as expected"
    cache_dir = tmp_path / "cache"
    extmods = tmp_path / "extmods"
    opts = {
        "id_function": {"grains.get": {"key": "osfinger"}},
        "cachedir": str(cache_dir),
        "extension_modules": str(extmods),
        "grains": {"osfinger": "meh"},
        "optimization_order": [0],
    }
    ret = salt.config.call_id_function(opts)
    assert ret == "meh"



"""
unit tests for salt.cache
"""


import salt.cache
import salt.payload
from tests.support.mock import patch

# import integration
from tests.support.unit import TestCase


class CacheFunctionsTest(TestCase):
    """
    Validate the cache package functions.
    """

    def setUp(self):
        self.opts = {
            "cache": "localfs",
            "memcache_expire_seconds": 0,
            "memcache_max_items": 0,
            "memcache_full_cleanup": False,
            "memcache_debug": False,
        }

    def test_factory_cache(self):
        ret = salt.cache.factory(self.opts)
        self.assertIsInstance(ret, salt.cache.Cache)

    def test_factory_memcache(self):
        self.opts["memcache_expire_seconds"] = 10
        ret = salt.cache.factory(self.opts)
        self.assertIsInstance(ret, salt.cache.MemCache)


class MemCacheTest(TestCase):
    """
    Validate Cache class methods
    """

    @patch("salt.payload.Serial")
    def setUp(self, serial_mock):  # pylint: disable=W0221
        salt.cache.MemCache.data = {}
        self.opts = {
            "cache": "fake_driver",
            "memcache_expire_seconds": 10,
            "memcache_max_items": 3,
            "memcache_full_cleanup": False,
            "memcache_debug": False,
        }
        self.cache = salt.cache.factory(self.opts)

    @patch("salt.cache.Cache.fetch", return_value="fake_data")
    @patch("salt.loader.cache", return_value={})
    def test_fetch(self, loader_mock, cache_fetch_mock):
        # Fetch value, it will be kept in cache.
        with patch("time.time", return_value=0):
            ret = self.cache.fetch("bank", "key")
        self.assertEqual(ret, "fake_data")
        self.assertDictEqual(
            salt.cache.MemCache.data,
            {"fake_driver": {("bank", "key"): [0, "fake_data"]}},
        )
        cache_fetch_mock.assert_called_once_with("bank", "key")
        cache_fetch_mock.reset_mock()

        # Fetch again, cached value is used, time updated.
        with patch("time.time", return_value=1):
            ret = self.cache.fetch("bank", "key")
        self.assertEqual(ret, "fake_data")
        self.assertDictEqual(
            salt.cache.MemCache.data,
            {"fake_driver": {("bank", "key"): [1, "fake_data"]}},
        )
        cache_fetch_mock.assert_not_called()

        # Fetch after expire
        with patch("time.time", return_value=12):
            ret = self.cache.fetch("bank", "key")
        self.assertEqual(ret, "fake_data")
        self.assertDictEqual(
            salt.cache.MemCache.data,
            {"fake_driver": {("bank", "key"): [12, "fake_data"]}},
        )
        cache_fetch_mock.assert_called_once_with("bank", "key")
        cache_fetch_mock.reset_mock()

    @patch("salt.cache.Cache.store")
    @patch("salt.loader.cache", return_value={})
    def test_store(self, loader_mock, cache_store_mock):
        # Fetch value, it will be kept in cache.
        with patch("time.time", return_value=0):
            self.cache.store("bank", "key", "fake_data")
        self.assertDictEqual(
            salt.cache.MemCache.data,
            {"fake_driver": {("bank", "key"): [0, "fake_data"]}},
        )
        cache_store_mock.assert_called_once_with("bank", "key", "fake_data")
        cache_store_mock.reset_mock()

        # Store another value.
        with patch("time.time", return_value=1):
            self.cache.store("bank", "key2", "fake_data2")
        self.assertDictEqual(
            salt.cache.MemCache.data,
            {
                "fake_driver": {
                    ("bank", "key"): [0, "fake_data"],
                    ("bank", "key2"): [1, "fake_data2"],
                }
            },
        )
        cache_store_mock.assert_called_once_with("bank", "key2", "fake_data2")

    @patch("salt.cache.Cache.store")
    @patch("salt.cache.Cache.flush")
    @patch("salt.loader.cache", return_value={})
    def test_flush(self, loader_mock, cache_flush_mock, cache_store_mock):
        # Flush non-existing bank
        self.cache.flush("bank")
        self.assertDictEqual(salt.cache.MemCache.data, {"fake_driver": {}})
        cache_flush_mock.assert_called_once_with("bank", None)
        cache_flush_mock.reset_mock()
        # Flush non-existing key
        self.cache.flush("bank", "key")
        self.assertDictEqual(salt.cache.MemCache.data, {"fake_driver": {}})
        cache_flush_mock.assert_called_once_with("bank", "key")
        cache_flush_mock.reset_mock()
        # Flush existing key
        with patch("time.time", return_value=0):
            self.cache.store("bank", "key", "fake_data")
        self.assertEqual(
            salt.cache.MemCache.data["fake_driver"][("bank", "key")], [0, "fake_data"]
        )
        self.assertDictEqual(
            salt.cache.MemCache.data,
            {"fake_driver": {("bank", "key"): [0, "fake_data"]}},
        )
        self.cache.flush("bank", "key")
        self.assertDictEqual(salt.cache.MemCache.data, {"fake_driver": {}})
        cache_flush_mock.assert_called_once_with("bank", "key")
        cache_flush_mock.reset_mock()

    @patch("salt.cache.Cache.store")
    @patch("salt.loader.cache", return_value={})
    def test_max_items(self, loader_mock, cache_store_mock):
        # Put MAX=3 values
        with patch("time.time", return_value=0):
            self.cache.store("bank1", "key1", "fake_data11")
        with patch("time.time", return_value=1):
            self.cache.store("bank1", "key2", "fake_data12")
        with patch("time.time", return_value=2):
            self.cache.store("bank2", "key1", "fake_data21")
        self.assertDictEqual(
            salt.cache.MemCache.data["fake_driver"],
            {
                ("bank1", "key1"): [0, "fake_data11"],
                ("bank1", "key2"): [1, "fake_data12"],
                ("bank2", "key1"): [2, "fake_data21"],
            },
        )
        # Put one more and check the oldest was removed
        with patch("time.time", return_value=3):
            self.cache.store("bank2", "key2", "fake_data22")
        self.assertDictEqual(
            salt.cache.MemCache.data["fake_driver"],
            {
                ("bank1", "key2"): [1, "fake_data12"],
                ("bank2", "key1"): [2, "fake_data21"],
                ("bank2", "key2"): [3, "fake_data22"],
            },
        )

    @patch("salt.cache.Cache.store")
    @patch("salt.loader.cache", return_value={})
    def test_full_cleanup(self, loader_mock, cache_store_mock):
        # Enable full cleanup
        self.cache.cleanup = True
        # Put MAX=3 values
        with patch("time.time", return_value=0):
            self.cache.store("bank1", "key1", "fake_data11")
        with patch("time.time", return_value=1):
            self.cache.store("bank1", "key2", "fake_data12")
        with patch("time.time", return_value=2):
            self.cache.store("bank2", "key1", "fake_data21")
        self.assertDictEqual(
            salt.cache.MemCache.data["fake_driver"],
            {
                ("bank1", "key1"): [0, "fake_data11"],
                ("bank1", "key2"): [1, "fake_data12"],
                ("bank2", "key1"): [2, "fake_data21"],
            },
        )
        # Put one more and check all expired was removed
        with patch("time.time", return_value=12):
            self.cache.store("bank2", "key2", "fake_data22")
        self.assertDictEqual(
            salt.cache.MemCache.data["fake_driver"],
            {
                ("bank2", "key1"): [2, "fake_data21"],
                ("bank2", "key2"): [12, "fake_data22"],
            },
        )

    @patch("salt.cache.Cache.fetch", return_value="fake_data")
    @patch("salt.loader.cache", return_value={})
    def test_fetch_debug(self, loader_mock, cache_fetch_mock):
        # Recreate cache with debug enabled
        self.opts["memcache_debug"] = True
        self.cache = salt.cache.factory(self.opts)

        # Fetch 2 values (no cache hit)
        with patch("time.time", return_value=0):
            ret = self.cache.fetch("bank", "key1")
        with patch("time.time", return_value=1):
            ret = self.cache.fetch("bank", "key2")
        # Fetch 3 times (cache hit)
        with patch("time.time", return_value=2):
            ret = self.cache.fetch("bank", "key2")
        with patch("time.time", return_value=3):
            ret = self.cache.fetch("bank", "key1")
        with patch("time.time", return_value=4):
            ret = self.cache.fetch("bank", "key1")
        # Fetch an expired value (no cache hit)
        with patch("time.time", return_value=13):
            ret = self.cache.fetch("bank", "key2")

        # Check debug data
        self.assertEqual(self.cache.call, 6)
        self.assertEqual(self.cache.hit, 3)



import logging

import pytest
from tests.support.case import ModuleCase, ShellCase

log = logging.getLogger(__name__)


class ExecutorTest(ModuleCase, ShellCase):
    def setup(self):
        self.run_function("saltutil.sync_all")

    @pytest.mark.slow_test
    def test_executor(self):
        """
        test that dunders are set
        """
        data = self.run_call("test.arg --module-executors=arg")
        self.assertIn("test.arg fired", "".join(data))

    @pytest.mark.slow_test
    def test_executor_with_multijob(self):
        """
        test that executor is fired when sending a multifunction job
        """
        data = self.run_salt("'*' test.arg,test.arg foo,bar --module-executors=arg")
        self.assertIn("test.arg fired", "".join(data))



import os
from tempfile import NamedTemporaryFile

import salt.modules.timezone as timezone
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase, skipIf

GET_ZONE_FILE = "salt.modules.timezone._get_zone_file"
GET_LOCALTIME_PATH = "salt.modules.timezone._get_localtime_path"


class TimezoneTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {timezone: {"__grains__": {"os_family": "Ubuntu"}}}

    def setUp(self):
        self.tempfiles = []

    def tearDown(self):
        for tempfile in self.tempfiles:
            try:
                os.remove(tempfile.name)
            except OSError:
                pass
        del self.tempfiles

    def test_zone_compare_equal(self):
        etc_localtime = self.create_tempfile_with_contents("a")
        zone_path = self.create_tempfile_with_contents("a")

        with patch(GET_ZONE_FILE, lambda p: zone_path.name):
            with patch(GET_LOCALTIME_PATH, lambda: etc_localtime.name):

                self.assertTrue(timezone.zone_compare("foo"))

    def test_zone_compare_nonexistent(self):
        etc_localtime = self.create_tempfile_with_contents("a")

        with patch(GET_ZONE_FILE, lambda p: "/foopath/nonexistent"):
            with patch(GET_LOCALTIME_PATH, lambda: etc_localtime.name):

                self.assertRaises(SaltInvocationError, timezone.zone_compare, "foo")

    def test_zone_compare_unequal(self):
        etc_localtime = self.create_tempfile_with_contents("a")
        zone_path = self.create_tempfile_with_contents("b")

        with patch(GET_ZONE_FILE, lambda p: zone_path.name):
            with patch(GET_LOCALTIME_PATH, lambda: etc_localtime.name):

                self.assertFalse(timezone.zone_compare("foo"))

    def test_missing_localtime(self):
        with patch(GET_ZONE_FILE, lambda p: "/nonexisting"):
            with patch(GET_LOCALTIME_PATH, lambda: "/also-missing"):
                self.assertRaises(CommandExecutionError, timezone.zone_compare, "foo")

    def create_tempfile_with_contents(self, contents):
        temp = NamedTemporaryFile(delete=False)
        temp.write(salt.utils.stringutils.to_bytes(contents))
        temp.close()
        self.tempfiles.append(temp)
        return temp


class TimezoneModuleTestCase(TestCase, LoaderModuleMockMixin):
    """
    Timezone test case
    """

    TEST_TZ = "UTC"

    def setup_loader_modules(self):
        return {
            timezone: {
                "__grains__": {"os": ""},
                "__salt__": {
                    "file.sed": MagicMock(),
                    "cmd.run": MagicMock(),
                    "cmd.retcode": MagicMock(return_value=0),
                },
            }
        }

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    def test_get_zone_centos(self):
        """
        Test CentOS is recognized
        :return:
        """
        with patch.dict(timezone.__grains__, {"os": "centos"}):
            with patch(
                "salt.modules.timezone._get_zone_etc_localtime",
                MagicMock(return_value=self.TEST_TZ),
            ):
                assert timezone.get_zone() == self.TEST_TZ

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    def test_get_zone_os_family_rh_suse(self):
        """
        Test RedHat and Suse are recognized
        :return:
        """
        for osfamily in ["RedHat", "Suse"]:
            with patch.dict(timezone.__grains__, {"os_family": [osfamily]}):
                with patch(
                    "salt.modules.timezone._get_zone_sysconfig",
                    MagicMock(return_value=self.TEST_TZ),
                ):
                    assert timezone.get_zone() == self.TEST_TZ

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    def test_get_zone_os_family_debian_gentoo(self):
        """
        Test Debian and Gentoo are recognized
        :return:
        """
        for osfamily in ["Debian", "Gentoo"]:
            with patch.dict(timezone.__grains__, {"os_family": [osfamily]}):
                with patch(
                    "salt.modules.timezone._get_zone_etc_timezone",
                    MagicMock(return_value=self.TEST_TZ),
                ):
                    assert timezone.get_zone() == self.TEST_TZ

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    def test_get_zone_os_family_allbsd_nilinuxrt_slackware(self):
        """
        Test *BSD, NILinuxRT and Slackware are recognized
        :return:
        """
        for osfamily in ["FreeBSD", "OpenBSD", "NetBSD", "NILinuxRT", "Slackware"]:
            with patch.dict(timezone.__grains__, {"os_family": osfamily}):
                with patch(
                    "salt.modules.timezone._get_zone_etc_localtime",
                    MagicMock(return_value=self.TEST_TZ),
                ):
                    assert timezone.get_zone() == self.TEST_TZ

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    def test_get_zone_os_family_slowlaris(self):
        """
        Test Slowlaris is recognized
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Solaris"]}):
            with patch(
                "salt.modules.timezone._get_zone_solaris",
                MagicMock(return_value=self.TEST_TZ),
            ):
                assert timezone.get_zone() == self.TEST_TZ

    @patch("salt.utils.path.which", MagicMock(return_value=False))
    def test_get_zone_os_family_aix(self):
        """
        Test IBM AIX is recognized
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["AIX"]}):
            with patch(
                "salt.modules.timezone._get_zone_aix",
                MagicMock(return_value=self.TEST_TZ),
            ):
                assert timezone.get_zone() == self.TEST_TZ

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_set_zone_os_family_nilinuxrt(self):
        """
        Test zone set on NILinuxRT
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["NILinuxRT"]}), patch.dict(
            timezone.__grains__, {"lsb_distrib_id": "nilrt"}
        ):
            assert timezone.set_zone(self.TEST_TZ)

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_set_zone_os_family_allbsd_slackware(self):
        """
        Test zone set on *BSD and Slackware
        :return:
        """
        for osfamily in ["FreeBSD", "OpenBSD", "NetBSD", "Slackware"]:
            with patch.dict(timezone.__grains__, {"os_family": osfamily}):
                assert timezone.set_zone(self.TEST_TZ)

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_set_zone_redhat(self):
        """
        Test zone set on RH series
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["RedHat"]}):
            assert timezone.set_zone(self.TEST_TZ)
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/sysconfig/clock", "^ZONE=.*", 'ZONE="UTC"')

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_set_zone_suse(self):
        """
        Test zone set on SUSE series
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Suse"]}):
            assert timezone.set_zone(self.TEST_TZ)
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/sysconfig/clock", "^TIMEZONE=.*", 'TIMEZONE="UTC"')

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_set_zone_gentoo(self):
        """
        Test zone set on Gentoo series
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Gentoo"]}):
            with patch("salt.utils.files.fopen", mock_open()) as m_open:
                assert timezone.set_zone(self.TEST_TZ)
                fh_ = m_open.filehandles["/etc/timezone"][0]
                assert fh_.call.args == ("/etc/timezone", "w"), fh_.call.args
                assert fh_.write_calls == ["UTC", "\n"], fh_.write_calls

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_set_zone_debian(self):
        """
        Test zone set on Debian series
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Debian"]}):
            with patch("salt.utils.files.fopen", mock_open()) as m_open:
                assert timezone.set_zone(self.TEST_TZ)
                fh_ = m_open.filehandles["/etc/timezone"][0]
                assert fh_.call.args == ("/etc/timezone", "w"), fh_.call.args
                assert fh_.write_calls == ["UTC", "\n"], fh_.write_calls

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=True))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_get_hwclock_timedate_utc(self):
        """
        Test get hwclock UTC/localtime
        :return:
        """
        with patch(
            "salt.modules.timezone._timedatectl",
            MagicMock(return_value={"stdout": "rtc in local tz"}),
        ):
            assert timezone.get_hwclock() == "UTC"
        with patch(
            "salt.modules.timezone._timedatectl",
            MagicMock(return_value={"stdout": "rtc in local tz:yes"}),
        ):
            assert timezone.get_hwclock() == "localtime"

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_get_hwclock_suse(self):
        """
        Test get hwclock on SUSE
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Suse"]}):
            timezone.get_hwclock()
            name, args, kwarg = timezone.__salt__["cmd.run"].mock_calls[0]
            assert args == (["tail", "-n", "1", "/etc/adjtime"],)
            assert kwarg == {"python_shell": False}

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_get_hwclock_redhat(self):
        """
        Test get hwclock on RedHat
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["RedHat"]}):
            timezone.get_hwclock()
            name, args, kwarg = timezone.__salt__["cmd.run"].mock_calls[0]
            assert args == (["tail", "-n", "1", "/etc/adjtime"],)
            assert kwarg == {"python_shell": False}

    def _test_get_hwclock_debian(
        self,
    ):  # TODO: Enable this when testing environment is working properly
        """
        Test get hwclock on Debian
        :return:
        """
        with patch("salt.utils.path.which", MagicMock(return_value=False)):
            with patch("os.path.exists", MagicMock(return_value=True)):
                with patch("os.unlink", MagicMock()):
                    with patch("os.symlink", MagicMock()):
                        with patch.dict(timezone.__grains__, {"os_family": ["Debian"]}):
                            timezone.get_hwclock()
                            name, args, kwarg = timezone.__salt__["cmd.run"].mock_calls[
                                0
                            ]
                            assert args == (["tail", "-n", "1", "/etc/adjtime"],)
                            assert kwarg == {"python_shell": False}

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_get_hwclock_solaris(self):
        """
        Test get hwclock on Solaris
        :return:
        """
        # Incomplete
        with patch.dict(timezone.__grains__, {"os_family": ["Solaris"]}):
            assert timezone.get_hwclock() == "UTC"
            with patch("salt.utils.files.fopen", mock_open()):
                assert timezone.get_hwclock() == "localtime"

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_get_hwclock_aix(self):
        """
        Test get hwclock on AIX
        :return:
        """
        # Incomplete
        hwclock = "localtime"
        if not os.path.isfile("/etc/environment"):
            hwclock = "UTC"
        with patch.dict(timezone.__grains__, {"os_family": ["AIX"]}):
            assert timezone.get_hwclock() == hwclock

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_get_hwclock_slackware_with_adjtime(self):
        """
        Test get hwclock on Slackware with /etc/adjtime present
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Slackware"]}):
            timezone.get_hwclock()
            name, args, kwarg = timezone.__salt__["cmd.run"].mock_calls[0]
            assert args == (["tail", "-n", "1", "/etc/adjtime"],)
            assert kwarg == {"python_shell": False}

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=False))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_get_hwclock_slackware_without_adjtime(self):
        """
        Test get hwclock on Slackware without /etc/adjtime present
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Slackware"]}):
            with patch("salt.utils.files.fopen", mock_open(read_data="UTC")):
                assert timezone.get_hwclock() == "UTC"
            with patch("salt.utils.files.fopen", mock_open(read_data="localtime")):
                assert timezone.get_hwclock() == "localtime"

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=True))
    def test_set_hwclock_timedatectl(self):
        """
        Test set hwclock with timedatectl
        :return:
        """
        timezone.set_hwclock("UTC")
        name, args, kwargs = timezone.__salt__["cmd.retcode"].mock_calls[0]
        assert args == (["timedatectl", "set-local-rtc", "false"],)

        timezone.set_hwclock("localtime")
        name, args, kwargs = timezone.__salt__["cmd.retcode"].mock_calls[1]
        assert args == (["timedatectl", "set-local-rtc", "true"],)

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    def test_set_hwclock_aix_nilinuxrt(self):
        """
        Test set hwclock on AIX and NILinuxRT
        :return:
        """
        for osfamily in ["AIX", "NILinuxRT"]:
            with patch.dict(timezone.__grains__, {"os_family": osfamily}):
                with self.assertRaises(SaltInvocationError):
                    assert timezone.set_hwclock("forty two")
                assert timezone.set_hwclock("UTC")

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    @patch("salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE"))
    def test_set_hwclock_solaris(self):
        """
        Test set hwclock on Solaris
        :return:
        """
        with patch.dict(
            timezone.__grains__, {"os_family": ["Solaris"], "cpuarch": "x86"}
        ):
            with self.assertRaises(SaltInvocationError):
                assert timezone.set_hwclock("forty two")
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["cmd.retcode"].mock_calls[0]
            assert args == (["rtc", "-z", "GMT"],)
            assert kwargs == {"python_shell": False}

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    @patch("salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE"))
    def test_set_hwclock_arch(self):
        """
        Test set hwclock on arch
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Arch"]}):
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["cmd.retcode"].mock_calls[0]
            assert args == (["timezonectl", "set-local-rtc", "false"],)
            assert kwargs == {"python_shell": False}

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    @patch("salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE"))
    def test_set_hwclock_redhat(self):
        """
        Test set hwclock on RedHat
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["RedHat"]}):
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/sysconfig/clock", "^ZONE=.*", 'ZONE="TEST_TIMEZONE"')

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    @patch("salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE"))
    def test_set_hwclock_suse(self):
        """
        Test set hwclock on SUSE
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Suse"]}):
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == (
                "/etc/sysconfig/clock",
                "^TIMEZONE=.*",
                'TIMEZONE="TEST_TIMEZONE"',
            )

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    @patch("salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE"))
    def test_set_hwclock_debian(self):
        """
        Test set hwclock on Debian
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Debian"]}):
            assert timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/default/rcS", "^UTC=.*", "UTC=yes")

            assert timezone.set_hwclock("localtime")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[1]
            assert args == ("/etc/default/rcS", "^UTC=.*", "UTC=no")

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    @patch("salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE"))
    def test_set_hwclock_gentoo(self):
        """
        Test set hwclock on Gentoo
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Gentoo"]}):
            with self.assertRaises(SaltInvocationError):
                timezone.set_hwclock("forty two")

            timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/conf.d/hwclock", "^clock=.*", 'clock="UTC"')

            timezone.set_hwclock("localtime")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[1]
            assert args == ("/etc/conf.d/hwclock", "^clock=.*", 'clock="local"')

    @skipIf(salt.utils.platform.is_windows(), "os.symlink not available in Windows")
    @patch("salt.utils.path.which", MagicMock(return_value=False))
    @patch("os.path.exists", MagicMock(return_value=True))
    @patch("os.unlink", MagicMock())
    @patch("os.symlink", MagicMock())
    @patch("salt.modules.timezone.get_zone", MagicMock(return_value="TEST_TIMEZONE"))
    def test_set_hwclock_slackware(self):
        """
        Test set hwclock on Slackware
        :return:
        """
        with patch.dict(timezone.__grains__, {"os_family": ["Slackware"]}):
            with self.assertRaises(SaltInvocationError):
                timezone.set_hwclock("forty two")

            timezone.set_hwclock("UTC")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[0]
            assert args == ("/etc/hardwareclock", "^(UTC|localtime)", "UTC")

            timezone.set_hwclock("localtime")
            name, args, kwargs = timezone.__salt__["file.sed"].mock_calls[1]
            assert args == ("/etc/hardwareclock", "^(UTC|localtime)", "localtime")



"""
Make api awesomeness
"""


import copy
import inspect
import logging
import os

import salt.auth
import salt.client
import salt.client.ssh.client
import salt.config
import salt.daemons.masterapi
import salt.exceptions
import salt.log  # pylint: disable=W0611
import salt.runner
import salt.syspaths
import salt.utils.args
import salt.utils.minions
import salt.wheel
from salt.defaults import DEFAULT_TARGET_DELIM

log = logging.getLogger(__name__)


class NetapiClient:
    """
    Provide a uniform method of accessing the various client interfaces in Salt
    in the form of low-data data structures. For example:

    >>> client = NetapiClient(__opts__)
    >>> lowstate = {'client': 'local', 'tgt': '*', 'fun': 'test.ping', 'arg': ''}
    >>> client.run(lowstate)
    """

    def __init__(self, opts):
        self.opts = opts
        apiopts = copy.deepcopy(self.opts)
        apiopts["enable_ssh_minions"] = True
        apiopts["cachedir"] = os.path.join(opts["cachedir"], "saltapi")
        if not os.path.exists(apiopts["cachedir"]):
            os.makedirs(apiopts["cachedir"])
        self.resolver = salt.auth.Resolver(apiopts)
        self.loadauth = salt.auth.LoadAuth(apiopts)
        self.key = salt.daemons.masterapi.access_keys(apiopts)
        self.ckminions = salt.utils.minions.CkMinions(apiopts)

    def _is_master_running(self):
        """
        Perform a lightweight check to see if the master daemon is running

        Note, this will return an invalid success if the master crashed or was
        not shut down cleanly.
        """
        # Windows doesn't have IPC. Assume the master is running.
        # At worse, it will error 500.
        if salt.utils.platform.is_windows():
            return True

        if self.opts["transport"] == "tcp":
            ipc_file = "publish_pull.ipc"
        else:
            ipc_file = "workers.ipc"
        return os.path.exists(os.path.join(self.opts["sock_dir"], ipc_file))

    def _prep_auth_info(self, clear_load):
        sensitive_load_keys = []
        key = None
        if "token" in clear_load:
            auth_type = "token"
            err_name = "TokenAuthenticationError"
            sensitive_load_keys = ["token"]
            return auth_type, err_name, key, sensitive_load_keys
        elif "eauth" in clear_load:
            auth_type = "eauth"
            err_name = "EauthAuthenticationError"
            sensitive_load_keys = ["username", "password"]
            return auth_type, err_name, key, sensitive_load_keys
        raise salt.exceptions.EauthAuthenticationError(
            "No authentication credentials given"
        )

    def _authorize_ssh(self, low):
        auth_type, err_name, key, sensitive_load_keys = self._prep_auth_info(low)
        auth_check = self.loadauth.check_authentication(low, auth_type, key=key)
        auth_list = auth_check.get("auth_list", [])
        error = auth_check.get("error")
        if error:
            raise salt.exceptions.EauthAuthenticationError(error)
        delimiter = low.get("kwargs", {}).get("delimiter", DEFAULT_TARGET_DELIM)
        _res = self.ckminions.check_minions(
            low["tgt"], low.get("tgt_type", "glob"), delimiter
        )
        minions = _res.get("minions", list())
        missing = _res.get("missing", list())
        authorized = self.ckminions.auth_check(
            auth_list,
            low["fun"],
            low.get("arg", []),
            low["tgt"],
            low.get("tgt_type", "glob"),
            minions=minions,
        )
        if not authorized:
            raise salt.exceptions.EauthAuthenticationError(
                "Authorization error occurred."
            )

    def run(self, low):
        """
        Execute the specified function in the specified client by passing the
        lowstate
        """
        # Eauth currently requires a running daemon and commands run through
        # this method require eauth so perform a quick check to raise a
        # more meaningful error.
        if not self._is_master_running():
            raise salt.exceptions.SaltDaemonNotRunning("Salt Master is not available.")

        if low.get("client") not in CLIENTS:
            raise salt.exceptions.SaltInvocationError(
                "Invalid client specified: '{}'".format(low.get("client"))
            )

        if not ("token" in low or "eauth" in low):
            raise salt.exceptions.EauthAuthenticationError(
                "No authentication credentials given"
            )

        if low.get("raw_shell") and not self.opts.get("netapi_allow_raw_shell"):
            raise salt.exceptions.EauthAuthenticationError(
                "Raw shell option not allowed."
            )

        if low["client"] == "ssh":
            self._authorize_ssh(low)

        l_fun = getattr(self, low["client"])
        f_call = salt.utils.args.format_call(l_fun, low)
        return l_fun(*f_call.get("args", ()), **f_call.get("kwargs", {}))

    def local_async(self, *args, **kwargs):
        """
        Run :ref:`execution modules <all-salt.modules>` asynchronously

        Wraps :py:meth:`salt.client.LocalClient.run_job`.

        :return: job ID
        """
        with salt.client.get_local_client(mopts=self.opts) as client:
            return client.run_job(*args, **kwargs)

    def local(self, *args, **kwargs):
        """
        Run :ref:`execution modules <all-salt.modules>` synchronously

        See :py:meth:`salt.client.LocalClient.cmd` for all available
        parameters.

        Sends a command from the master to the targeted minions. This is the
        same interface that Salt's own CLI uses. Note the ``arg`` and ``kwarg``
        parameters are sent down to the minion(s) and the given function,
        ``fun``, is called with those parameters.

        :return: Returns the result from the execution module
        """
        with salt.client.get_local_client(mopts=self.opts) as client:
            return client.cmd(*args, **kwargs)

    def local_subset(self, *args, **kwargs):
        """
        Run :ref:`execution modules <all-salt.modules>` against subsets of minions

        .. versionadded:: 2016.3.0

        Wraps :py:meth:`salt.client.LocalClient.cmd_subset`
        """
        with salt.client.get_local_client(mopts=self.opts) as client:
            return client.cmd_subset(*args, **kwargs)

    def local_batch(self, *args, **kwargs):
        """
        Run :ref:`execution modules <all-salt.modules>` against batches of minions

        .. versionadded:: 0.8.4

        Wraps :py:meth:`salt.client.LocalClient.cmd_batch`

        :return: Returns the result from the exeuction module for each batch of
            returns
        """
        with salt.client.get_local_client(mopts=self.opts) as client:
            return client.cmd_batch(*args, **kwargs)

    def ssh(self, *args, **kwargs):
        """
        Run salt-ssh commands synchronously

        Wraps :py:meth:`salt.client.ssh.client.SSHClient.cmd_sync`.

        :return: Returns the result from the salt-ssh command
        """
        with salt.client.ssh.client.SSHClient(
            mopts=self.opts, disable_custom_roster=True
        ) as client:
            return client.cmd_sync(kwargs)

    def runner(self, fun, timeout=None, full_return=False, **kwargs):
        """
        Run `runner modules <all-salt.runners>` synchronously

        Wraps :py:meth:`salt.runner.RunnerClient.cmd_sync`.

        Note that runner functions must be called using keyword arguments.
        Positional arguments are not supported.

        :return: Returns the result from the runner module
        """
        kwargs["fun"] = fun
        runner = salt.runner.RunnerClient(self.opts)
        return runner.cmd_sync(kwargs, timeout=timeout, full_return=full_return)

    def runner_async(self, fun, **kwargs):
        """
        Run `runner modules <all-salt.runners>` asynchronously

        Wraps :py:meth:`salt.runner.RunnerClient.cmd_async`.

        Note that runner functions must be called using keyword arguments.
        Positional arguments are not supported.

        :return: event data and a job ID for the executed function.
        """
        kwargs["fun"] = fun
        runner = salt.runner.RunnerClient(self.opts)
        return runner.cmd_async(kwargs)

    def wheel(self, fun, **kwargs):
        """
        Run :ref:`wheel modules <all-salt.wheel>` synchronously

        Wraps :py:meth:`salt.wheel.WheelClient.master_call`.

        Note that wheel functions must be called using keyword arguments.
        Positional arguments are not supported.

        :return: Returns the result from the wheel module
        """
        kwargs["fun"] = fun
        wheel = salt.wheel.WheelClient(self.opts)
        return wheel.cmd_sync(kwargs)

    def wheel_async(self, fun, **kwargs):
        """
        Run :ref:`wheel modules <all-salt.wheel>` asynchronously

        Wraps :py:meth:`salt.wheel.WheelClient.master_call`.

        Note that wheel functions must be called using keyword arguments.
        Positional arguments are not supported.

        :return: Returns the result from the wheel module
        """
        kwargs["fun"] = fun
        wheel = salt.wheel.WheelClient(self.opts)
        return wheel.cmd_async(kwargs)


CLIENTS = [
    name
    for name, _ in inspect.getmembers(NetapiClient, predicate=None)
    if not (name == "run" or name.startswith("_"))
]



"""
Process Management
==================

Ensure a process matching a given pattern is absent.

.. code-block:: yaml

    httpd-absent:
      process.absent:
        - name: apache2
"""


def __virtual__():
    if "ps.pkill" in __salt__:
        return True
    return (False, "ps module could not be loaded")


def absent(name, user=None, signal=None):
    """
    Ensures that the named command is not running.

    name
        The pattern to match.

    user
        The user to which the process belongs

    signal
        Signal to send to the process(es).
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if __opts__["test"]:
        running = __salt__["ps.pgrep"](name, user=user)
        ret["result"] = None
        if running:
            ret["comment"] = "{} processes will be killed".format(len(running))
        else:
            ret["comment"] = "No matching processes running"
        return ret

    if signal:
        status = __salt__["ps.pkill"](name, user=user, signal=signal, full=True)
    else:
        status = __salt__["ps.pkill"](name, user=user, full=True)

    ret["result"] = True
    if status:
        ret["comment"] = "Killed {} processes".format(len(status["killed"]))
        ret["changes"] = status
    else:
        ret["comment"] = "No matching processes running"
    return ret



"""
unit tests for the stalekey engine
"""

import logging

import salt.engines.stalekey as stalekey
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class MockWheel:
    def __init__(self, *args, **kwargs):
        pass

    def cmd(self, *args, **kwargs):
        return True


class EngineStalekeyTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.engine.stalekeys
    """

    def setup_loader_modules(self):
        return {stalekey: {}}

    def test__delete_keys(self):
        """
        Test to ensure the _delete_keys function deletes multiple keys
        """
        with patch("salt.wheel.WheelClient", MockWheel):
            stale_keys = ["minion1", "minion2"]
            minions = {
                "minion1": 1601430462.5281658,
                "minion2": 1601430462.5281658,
            }

            ret = stalekey._delete_keys(stale_keys, minions)
            self.assertEqual(ret, {})

            stale_keys = ["minion1"]
            minions = {
                "minion1": 1601430462.5281658,
                "minion2": 1601430462.5281658,
            }

            ret = stalekey._delete_keys(stale_keys, minions)
            self.assertEqual(ret, {"minion2": 1601430462.5281658})

    def test__read_presence(self):
        """
        Test for _read_presence returning False for no error and minions presence data
        """
        presence_data = {b"minion": 1601477127.532849}
        expected = (False, {"minion": 1601477127.532849})
        with patch("os.path.exists", return_value=True):
            with patch("salt.utils.files.fopen", mock_open()):
                with patch(
                    "salt.utils.msgpack.load", MagicMock(return_value=presence_data)
                ):
                    ret = stalekey._read_presence("presence_file")
                    self.assertEqual(ret, expected)

    def test__write_presence(self):
        """
        Test for _write_presence returning False, meaning no error has occured
        """
        expected = False
        minions = {
            "minion1": 1601430462.5281658,
            "minion2": 1601430462.5281658,
        }
        with patch("salt.utils.files.fopen", mock_open()):
            ret = stalekey._write_presence("presence_file", minions)
            self.assertEqual(ret, expected)
